import warnings

from sympy import Rational as frac
from sympy import sqrt

from .. import cn
from ..helpers import book
from ._albrecht_collatz import albrecht_collatz_1 as stroud_c2_3_4
from ._albrecht_collatz import albrecht_collatz_2 as stroud_c2_5_1
from ._albrecht_collatz import albrecht_collatz_3 as stroud_c2_5_2
from ._albrecht_collatz import albrecht_collatz_4 as stroud_c2_5_6
from ._burnside import burnside as stroud_c2_5_3
from ._helpers import C2Scheme, register
from ._irwin import irwin_1 as stroud_c2_3_5
from ._irwin import irwin_2 as stroud_c2_5_7
from ._maxwell import maxwell as stroud_c2_7_3
from ._meister import meister as stroud_c2_7_6
from ._miller import miller as stroud_c2_1_2
from ._phillips import phillips as stroud_c2_7_2
from ._rabinowitz_richter import rabinowitz_richter_1 as stroud_c2_9_1
from ._rabinowitz_richter import rabinowitz_richter_2 as stroud_c2_11_1
from ._rabinowitz_richter import rabinowitz_richter_3 as stroud_c2_11_2
from ._rabinowitz_richter import rabinowitz_richter_4 as stroud_c2_13_1
from ._rabinowitz_richter import rabinowitz_richter_5 as stroud_c2_15_1
from ._rabinowitz_richter import rabinowitz_richter_6 as stroud_c2_15_2
from ._tyler import tyler_1 as stroud_c2_5_5
from ._tyler import tyler_2 as stroud_c2_7_1
from ._tyler import tyler_3 as stroud_c2_7_5

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_c2_1_1():
    # product trapezoidal
    d = {"d4_aa": [[frac(1, 4)], [1]]}
    return C2Scheme("Stroud C2 1-1", d, 1, source)


def stroud_c2_3_1():
    # ERR misprint in Stroud: sqrt(1/3) vs 1/3
    d = {"d4_aa": [[frac(1, 4)], [sqrt(frac(1, 3))]]}
    return C2Scheme("Stroud C2 3-1", d, 3, source)


def stroud_c2_3_2():
    return cn.ewing(2)


def stroud_c2_3_3():
    return cn.stroud_cn_3_6(2)


def stroud_c2_5_4():
    r = sqrt(frac(3, 5))
    d = {
        "zero2": [[frac(16, 81)]],
        "d4_a0": [[frac(10, 81)], [r]],
        "d4_aa": [[frac(25, 324)], [r]],
    }
    return C2Scheme("Stroud C2 5-4", d, 5, source)


def stroud_c2_7_4():
    # product Gauss 7
    r, s = (sqrt((15 - i * 2 * sqrt(30)) / 35) for i in [+1, -1])

    B1, B2 = ((59 + i * 6 * sqrt(30)) / 864 for i in [+1, -1])
    B3 = frac(49, 864)

    r = sqrt(frac(3, 5))

    d = {"d4_aa": [[B1, B2], [r, s]], "d4_ab": [[B3], [r], [s]]}

    # TODO fix
    warnings.warn("Formula only has degree 1!")
    return C2Scheme("Stroud C2 7-4", d, 1, source)


register(
    [
        stroud_c2_1_1,
        stroud_c2_1_2,
        stroud_c2_3_1,
        stroud_c2_3_2,
        stroud_c2_3_3,
        stroud_c2_3_4,
        stroud_c2_3_5,
        stroud_c2_5_1,
        stroud_c2_5_2,
        stroud_c2_5_3,
        stroud_c2_5_4,
        stroud_c2_5_5,
        stroud_c2_5_6,
        stroud_c2_5_7,
        stroud_c2_7_1,
        stroud_c2_7_2,
        stroud_c2_7_3,
        stroud_c2_7_4,
        stroud_c2_7_5,
        stroud_c2_7_6,
        stroud_c2_9_1,
        stroud_c2_11_1,
        stroud_c2_11_2,
        stroud_c2_13_1,
        stroud_c2_15_1,
        stroud_c2_15_2,
    ]
)



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, rd, untangle
from ._helpers import TnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Some approximate integration formulas of degree 3 for an n-dimensional simplex",
    journal="Numerische Mathematik",
    month="nov",
    year="1966",
    volume="9",
    number="1",
    pages="38–45",
    url="https://doi.org/10.1007/BF02165227",
)


def stroud_1966_1(n):
    degree = 3
    sqrt4n13 = sqrt(4 * n + 13)

    r = (2 * n + 5 - sqrt4n13) / 2 / (n + 1) / (n + 3)
    s = 1 - n * r

    B = (1 - sqrt4n13) / 2 / (n + 1) / (n + 2) / (n + 3)
    C = (2 * n ** 2 + 10 * n + 11 + sqrt4n13) / 2 / (n + 1) / (n + 2) / (n + 3)

    data = [(B, rd(n + 1, [(1, 1)])), (C, rd(n + 1, [(r, n), (s, 1)]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-I", n, weights, points, degree, source)


def stroud_1966_2(n):
    degree = 3
    # r is a smallest real-valued root of a polynomial of degree 3
    rts = np.roots(
        [2 * (n - 2) * (n + 1) * (n + 3), -(5 * n ** 2 + 5 * n - 18), 4 * n, -1]
    )
    r = np.min([r.real for r in rts if abs(r.imag) < 1.0e-15])

    s = 1 - n * r
    t = 0.5

    B = (n - 2) / (1 - 2 * n * r ** 2 - 2 * (1 - n * r) ** 2) / (n + 1) / (n + 2)
    C = 2 * (1 / (n + 1) - B) / n

    data = [(B, rd(n + 1, [(r, n), (s, 1)])), (C, rd(n + 1, [(t, 2)]))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-II", n, weights, points, degree, source)


def stroud_1966_3(n):
    degree = 3
    assert n > 2

    r = frac(1, 2)
    s = frac(1, n)

    prod = (n + 1) * (n + 2) * (n + 3)
    B = frac(6 - n, prod)
    C = frac(8 * (n - 3), (n - 2) * prod)
    D = frac(n ** 3, (n - 2) * prod)

    data = [
        (B, rd(n + 1, [(1, 1)])),
        (C, rd(n + 1, [(r, 2)])),
        (D, rd(n + 1, [(s, n)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-III", n, weights, points, degree, source)


def stroud_1966_4(n):
    degree = 3
    assert n >= 3

    r = frac(1, n + 1)
    s = frac(1, 3)

    A = frac((n + 1) ** 2 * (n - 3), (n - 2) * (n + 2) * (n + 3))
    B = frac((9 - n), 2 * (n + 1) * (n + 2) * (n + 3))
    C = frac(27, (n - 2) * (n + 1) * (n + 2) * (n + 3))

    data = [
        (A, np.full((1, n + 1), r)),
        (B, rd(n + 1, [(1, 1)])),
        (C, rd(n + 1, [(s, 3)])),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-IV", n, weights, points, degree, source)


def stroud_1966_5(n):
    degree = 3
    r = frac(1, n)
    s = frac(1, 3)

    prod = (n + 1) * (n + 2) * (n + 3)
    A = frac(-(n ** 2) + 11 * n - 12, 2 * (n - 1) * prod)
    B = frac(n ** 3, (n - 1) * prod)
    C = frac(27, (n - 1) * prod)

    data = [
        (A, rd(n + 1, [(1, 1)])),
        (B, rd(n + 1, [(r, n)])),
        (C, rd(n + 1, [(s, 3)])),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-I", n, weights, points, degree, source)


def stroud_1966_6(n):
    degree = 3
    assert n >= 3
    assert n != 5

    r = frac(1, n + 1)
    s = frac(1, 3)
    t = frac(1, n - 2)

    prod = (n + 1) * (n + 2) * (n + 3)
    A = frac((3 - n) * (n - 12) * (n + 1) ** 2, 3 * (n - 2) * (n + 2) * (n + 3))
    B = frac(54 * (3 * n - 11), (n - 5) * (n - 2) * (n - 1) * prod)
    C = frac(2 * (n - 2) ** 2 * (n - 9), (n - 5) * (n - 1) * prod)

    data = [
        (A, np.full((1, n + 1), r)),
        (B, rd(n + 1, [(s, 3)])),
        (C, rd(n + 1, [(t, n - 2)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-VI", n, weights, points, degree, source)


def stroud_1966_7(n):
    degree = 3
    assert n >= 3
    assert n != 5

    s = frac(1, 3)
    t = frac(1, n - 2)

    prod = (n + 1) * (n + 2) * (n + 3)
    A = frac((12 - n), 2 * prod)
    B = frac(27 * (n - 7), (n - 5) * (n - 1) * prod)
    C = frac(6 * (n - 2) ** 2, (n - 5) * (n - 1) * prod)

    data = [
        (A, rd(n + 1, [(1, 1)])),
        (B, rd(n + 1, [(s, 3)])),
        (C, rd(n + 1, [(t, n - 2)])),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1966-VII", n, weights, points, degree, source)



import ndim

from ..cn._mcnamee_stenger import (
    _mcnamee_stenger_3,
    _mcnamee_stenger_5,
    _mcnamee_stenger_7,
    _mcnamee_stenger_9,
)
from ._helpers import Enr2Scheme


def integrator(n, k, symbolic):
    """Returns the integral of the polynomial given by the coefficients k over Enr2."""
    alpha = k + [0] * (n - len(k))
    return ndim.enr2.integrate_monomial(alpha, "physicists", symbolic=symbolic)


def mcnamee_stenger_3(n, symbolic=False):
    return Enr2Scheme(*_mcnamee_stenger_3(n, integrator, symbolic=symbolic))


def mcnamee_stenger_5(n, symbolic=False):
    return Enr2Scheme(*_mcnamee_stenger_5(n, integrator, symbolic=symbolic))


def mcnamee_stenger_7a(n, symbolic=False):
    return Enr2Scheme(
        *_mcnamee_stenger_7(n, integrator, False, symbolic=symbolic), tol=1.250e-14
    )


def mcnamee_stenger_7b(n, symbolic=False):
    return Enr2Scheme(*_mcnamee_stenger_7(n, integrator, True, symbolic=symbolic))


def mcnamee_stenger_9a(n, symbolic=False):
    return Enr2Scheme(*_mcnamee_stenger_9(n, integrator, False, symbolic=symbolic))


def mcnamee_stenger_9b(n, symbolic=False):
    return Enr2Scheme(*_mcnamee_stenger_9(n, integrator, True, symbolic=symbolic))



from __future__ import annotations

import json

import numpy as np

from ..helpers import expand_symmetries, plot_disks
from ..tn import TnScheme, get_vol, transform

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class T2Scheme(TnScheme):
    def __init__(
        self,
        name: str,
        symmetry_data,
        degree: int,
        source=None,
        tol: float = 1.0e-14,
        comments: list[str] | None = None,
        weight_factor: float | None = None,
    ):
        self.symmetry_data = symmetry_data

        points, weights = expand_symmetries(symmetry_data, dim=2)
        if weight_factor is not None:
            weights *= weight_factor

        assert points.shape[0] == 3, f"{name}, {points.shape}"
        super().__init__(name, 2, weights, points, degree, source, tol, comments)
        self.domain = "T2"

    def plot(
        self,
        triangle=np.array([[-0.5, 0.0], [+0.5, 0.0], [0, 0.5 * (np.sqrt(3))]]),
        show_axes: bool = False,
    ):
        """Shows the quadrature points on a given triangle. The size of the circles
        around the points coincides with their weights.
        """
        from matplotlib import pyplot as plt

        plt.plot(triangle[:, 0], triangle[:, 1], "-k")
        plt.plot(
            [triangle[-1, 0], triangle[0, 0]], [triangle[-1, 1], triangle[0, 1]], "-k"
        )

        if not show_axes:
            plt.gca().set_axis_off()

        transformed_pts = transform(self.points, triangle.T).T

        vol = get_vol(triangle)
        plot_disks(plt, transformed_pts, self.weights, vol)

        plt.axis("equal")


def _read(filepath, source):
    with open(filepath) as f:
        content = json.load(f)

    degree = content["degree"]
    name = content["name"]
    tol = content["test_tolerance"]
    d = content["data"]
    weight_factor = content["weight factor"] if "weight factor" in content else None
    return T2Scheme(name, d, degree, source, tol, weight_factor=weight_factor)


def _scheme_from_dict(content, source=None):
    points, weights = expand_symmetries(content["data"])

    if "weight factor" in content:
        weights *= content["weight factor"]

    return T2Scheme(
        content["name"],
        weights,
        points,
        degree=content["degree"],
        source=source,
        tol=content["test_tolerance"],
        comments=content["comments"] if "comments" in content else None,
    )


def get_good_scheme(degree: int) -> T2Scheme | None:
    if degree <= 50:
        return {
            0: schemes["centroid"],
            1: schemes["centroid"],
            2: schemes["hammer_marlowe_stroud_3"],
            3: schemes["strang_fix_cowper_04"],
            4: schemes["williams_shunn_jameson_3"],
            5: schemes["hammer_marlowe_stroud_5"],
            6: schemes["strang_fix_cowper_09"],
            7: schemes["laursen_gellert_11"],
            8: schemes["lyness_jespersen_15"],
            9: schemes["lyness_jespersen_18"],
            10: schemes["witherden_vincent_10"],
            11: schemes["xiao_gimbutas_11"],
            12: schemes["xiao_gimbutas_12"],
            13: schemes["xiao_gimbutas_13"],
            14: schemes["xiao_gimbutas_14"],
            15: schemes["witherden_vincent_15"],
            16: schemes["xiao_gimbutas_16"],
            17: schemes["xiao_gimbutas_17"],
            18: schemes["witherden_vincent_18"],
            19: schemes["dunavant_19"],
            20: schemes["xiao_gimbutas_20"],
            21: schemes["xiao_gimbutas_21"],
            22: schemes["xiao_gimbutas_22"],
            23: schemes["xiao_gimbutas_23"],
            24: schemes["xiao_gimbutas_24"],
            25: schemes["xiao_gimbutas_25"],
            26: schemes["xiao_gimbutas_26"],
            27: schemes["xiao_gimbutas_27"],
            28: schemes["xiao_gimbutas_28"],
            29: schemes["xiao_gimbutas_29"],
            30: schemes["xiao_gimbutas_30"],
            31: schemes["xiao_gimbutas_31"],
            32: schemes["xiao_gimbutas_32"],
            33: schemes["xiao_gimbutas_33"],
            34: schemes["xiao_gimbutas_34"],
            35: schemes["xiao_gimbutas_35"],
            36: schemes["xiao_gimbutas_36"],
            37: schemes["xiao_gimbutas_37"],
            38: schemes["xiao_gimbutas_38"],
            39: schemes["xiao_gimbutas_39"],
            40: schemes["xiao_gimbutas_40"],
            41: schemes["xiao_gimbutas_41"],
            42: schemes["xiao_gimbutas_42"],
            43: schemes["xiao_gimbutas_43"],
            44: schemes["xiao_gimbutas_44"],
            45: schemes["xiao_gimbutas_45"],
            46: schemes["xiao_gimbutas_46"],
            47: schemes["xiao_gimbutas_47"],
            48: schemes["xiao_gimbutas_48"],
            49: schemes["xiao_gimbutas_49"],
            50: schemes["xiao_gimbutas_50"],
        }[degree]()

    # degree > 50
    return None



from ..helpers import article

# ERR All schemes only have degree 1.
source = article(
    authors=["K. Sham Sunder", "R.A. Cookson"],
    title="Integration points for triangles and tetrahedrons obtained from the Gaussian quadrature points for a line",
    journal="Computers & Structures",
    year="1985",
    volume="21",
    number="5",
    pages="881-885",
    url="https://doi.org/10.1016/0045-7949%2885%2990198-1",
)


# def sunder_cookson_01():
#     weights, points = s3([1])
#     return T2Scheme("Sunder-Cookson 1", weights, points, 1, source)
#
#
# def sunder_cookson_02():
#     # ERR
#     warnings.warn("Sunder-Cookson claim degree 2, but the scheme is only degree 1.")
#     alpha = (1 - 1 / sqrt(3)) / 2
#     l2 = alpha / (alpha + 1)
#     weights, points = s2([1 / 3, l2])
#     return T2Scheme("Sunder-Cookson 2", weights, points, 1, source)
#
#
# def sunder_cookson_03():
#     # ERR
#     warnings.warn("Sunder-Cookson claim degree 3, but the scheme is only degree 1.")
#     weights, points = concat(s3(4 / 9), s2([5 / 27, 0.101286507323456]))
#     return T2Scheme("Sunder-Cookson 3", weights, points, 1, source)
#
#
# def sunder_cookson_04():
#     # ERR
#     warnings.warn("Sunder-Cookson claim degree 4, but the scheme is only degree 1.")
#     weights, points = concat(
#         s2(0.115951615045818, 0.064924047829079),
#         s2(0.217381718287515, 0.248125658963212),
#     )
#     return T2Scheme("Sunder-Cookson 4", weights, points, 1, source)
#
#
# def sunder_cookson_05():
#     # ERR
#     warnings.warn("Sunder-Cookson claim degree 5, but the scheme is only degree 1.")
#     weights, points = concat(
#         s3(64 / 225),
#         s2(0.078975628352063, 0.044808124460621),
#         s2(0.159542890166455, 0.187497434742173),
#     )
#     return T2Scheme("Sunder-Cookson 5", weights, points, 1, source)



import ndim
import numpy as np
from numpy import sqrt

from ..helpers import article, fsd, pm, untangle
from ._helpers import UnScheme

source = article(
    authors=["A.H. Stroud"],
    title="A Fifth Degree Integration Formula for the n-Simplex",
    journal="SIAM J. Numer. Anal.",
    volume="6",
    number="1",
    pages="90–98",
    url="https://doi.org/10.1137/0706009",
)


def stroud_1969(n):
    assert n > 2
    degree = 11

    sqrt3 = sqrt(3)

    t = sqrt(1 / n)
    r1, r2 = (sqrt((n + 6 - i * 4 * sqrt3) / (n ** 2 + 12 * n - 12)) for i in [+1, -1])
    s1, s2 = (
        sqrt((7 * n - 6 + i * 4 * (n - 1) * sqrt3) / (n ** 2 + 12 * n - 12))
        for i in [+1, -1]
    )
    u1, u2 = (sqrt((n + 12 + i * 8 * sqrt3) / (n ** 2 + 24 * n - 48)) for i in [+1, -1])
    v1, v2 = (
        sqrt((7 * n - 12 - i * 4 * (n - 2) * sqrt3) / (n ** 2 + 24 * n - 48))
        for i in [+1, -1]
    )

    # Solve linear equation system for x^k, k={0, 4, 6, 8, 10}, for the weights (the
    # same is done in Stroud's article).
    pts = [pm(n * [t]), fsd(n, (s1, 1), (r1, n - 1)), fsd(n, (s2, 1), (r2, n - 1))]
    k_range = [0, 4, 6]
    if n >= 4:
        pts.append(fsd(n, (v1, 2), (u1, n - 2)))
        k_range.append(8)
    if n >= 5:
        pts.append(fsd(n, (v2, 2), (u2, n - 2)))
        k_range.append(10)
    # TODO build the equation system from orthogonal polynomials
    b = [
        ndim.nsphere.integrate_monomial([k] + (n - 1) * [0], symbolic=False)
        for k in k_range
    ]
    A = [[sum(p[:, 0] ** k) for p in pts] for k in k_range]
    w = np.linalg.solve(A, b)

    data = [(w[k], pts[k]) for k in range(len(w))]

    points, weights = untangle(data)
    weights /= ndim.nsphere.volume(n)
    return UnScheme("Stroud 1969", n, weights, points, degree, source, 4.270e-14)



"""
Two of the schemes also appear in

P.C. Hammer, Arthur H. Stroud,
Numerical Evaluation of Multiple Integrals II,
Mathematical Tables and Other Aids to Computation.
Vol. 12, No. 64 (Oct., 1958), pp. 272-280,
<https://www.jstor.org/stable/2002370>
"""
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import T2Scheme, register

source = article(
    authors=["P.C. Hammer", "O.J. Marlowe", "A.H. Stroud"],
    title="Numerical Integration Over Simplexes and Cones",
    journal="Mathematical Tables and Other Aids to Computation",
    volume="10",
    number="55",
    month="jul",
    year="1956",
    pages="130-137",
    url="https://doi.org/10.1090/S0025-5718-1956-0086389-6",
)


def hammer_marlowe_stroud_1():
    d = {"centroid": [[1]]}
    return T2Scheme("Hammer-Marlowe-Stroud 1", d, 1, source, 7.850e-17)


def hammer_marlowe_stroud_2():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Hammer-Marlowe-Stroud 2", d, 2, source, 2.220e-16)


def hammer_marlowe_stroud_3():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 2)]]}
    return T2Scheme("Hammer-Marlowe-Stroud 3", d, 2, source, 3.074e-16)


def hammer_marlowe_stroud_4():
    d = {"centroid": [[-frac(9, 16)]], "d3_aa": [[frac(25, 48)], [frac(1, 5)]]}
    return T2Scheme("Hammer-Marlowe-Stroud 4", d, 3, source, 6.661e-16)


def hammer_marlowe_stroud_5():
    w1, w2 = ((155 - i * sqrt(15)) / 1200 for i in [+1, -1])
    x1, x2 = ((1 + i * sqrt(15)) / 7 for i in [+1, -1])
    b1 = (1 - x1) / 3
    b2 = (1 - x2) / 3
    d = {"centroid": [[frac(9, 40)]], "d3_aa": [[w1, w2], [b1, b2]]}
    return T2Scheme("Hammer-Marlowe-Stroud 5", d, 5, source, 2.776e-16)


register(
    [
        hammer_marlowe_stroud_1,
        hammer_marlowe_stroud_2,
        hammer_marlowe_stroud_3,
        hammer_marlowe_stroud_4,
        hammer_marlowe_stroud_5,
    ]
)



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, untangle
from ._helpers import Enr2Scheme

source = article(
    authors=["G.M. Phillips"],
    title="A survey of one-dimensional and multidimensional numerical integration",
    journal="Computer Physics Communications",
    volume="20",
    number="1",
    year="1980",
    pages="17-27",
    url="https://doi.org/10.1016/0010-4655%2880%2990102-2",
)


def phillips(n):
    lmbda = sqrt(frac(3, 2))
    w1 = frac(n ** 2 - 7 * n + 18, 18)
    w2 = frac(4 - n, 18)
    w3 = frac(1, 36)

    data = [
        (w1, np.full((1, n), 0)),
        (w2, fsd(n, (lmbda, 1))),
        (w3, fsd(n, (lmbda, 2))),
    ]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return Enr2Scheme("Phillips", n, weights, points, 5, source)



import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["Philip Rabinowitz", "Nira Richter"],
    title="Perfectly Symmetric Two-Dimensional Integration Formulas with Minimal Numbers of Points",
    journal="Mathematics of Computation",
    volume="23",
    number="108",
    month="oct",
    year="1969",
    pages="765-779",
    url="https://doi.org/10.1090/S0025-5718-1969-0258281-4",
)

this_dir = pathlib.Path(__file__).resolve().parent


def rabinowitz_richter_1():
    return _read(this_dir / "rabinowitz_richter_1.json", _source)


def rabinowitz_richter_2():
    return _read(this_dir / "rabinowitz_richter_2.json", _source)


def rabinowitz_richter_3():
    return _read(this_dir / "rabinowitz_richter_3.json", _source)


# ERR There's a misprint here somewhere.
# The sum of the weights is 406.4177935852262, not 6.283185307179584 as it should be.
# When replacing -.1010440929995067e+1 by -.1010440929995067e+3, this is fixed, but the
# scheme still only has order 3.
# We can better tackle this once we have orthogonal polynomials for E2r.
# TODO find out what's going wrong
# def rabinowitz_richter_4():
#    return _read(this_dir / "rabinowitz_richter_4.json", _source)


def rabinowitz_richter_5():
    return _read(this_dir / "rabinowitz_richter_5.json", _source)


register(
    [
        rabinowitz_richter_1,
        rabinowitz_richter_2,
        rabinowitz_richter_3,
        # rabinowitz_richter_4,
        rabinowitz_richter_5,
    ]
)



import math

import numpy as np
import sympy

from ..helpers import book, untangle
from ..u3 import _stroud as sphere_stroud
from ._ditkin import ditkin_1 as stroud_5_1
from ._ditkin import ditkin_2 as stroud_5_2
from ._ditkin import ditkin_3 as stroud_7_3
from ._hammer_stroud import hammer_stroud_11_3 as stroud_3_1
from ._hammer_stroud import hammer_stroud_15_3a as stroud_7_1a
from ._hammer_stroud import hammer_stroud_15_3b as stroud_7_1b
from ._helpers import S3Scheme, register
from ._mysovskih import mysovskih as stroud_7_2

_source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)

pi = sympy.pi
sqrt = np.vectorize(sympy.sqrt)


def stroud_7_4():
    # spherical product gauss
    # ENH Stroud only gives decimals, sophisticated guesswork gives the analytical
    # expressions.
    pm = np.array([+1, -1])

    # 0.9061798459, 0.5384691101
    alpha, beta = sqrt((35 + pm * 2 * sqrt(70)) / 63)
    rho = np.array([-alpha, -beta, beta, alpha])

    # 0.8611363116, 0.3399810436
    alpha, beta = sqrt((15 + pm * 2 * sqrt(30)) / 35)
    u = np.array([-alpha, -beta, beta, alpha])

    # 0.9238795325, 0.3826834324
    alpha, beta = sqrt((2 + pm * sqrt(2)) / 4)
    v = np.array([-alpha, -beta, beta, alpha])

    # 0.1945553342, 0.1387779991
    alpha, beta = (50 + pm * sqrt(70)) / 300
    A = np.array([alpha, beta, beta, alpha])

    # 0.3478548451, 0.6521451549
    alpha, beta = (18 - pm * sqrt(30)) / 36
    B = np.array([alpha, beta, beta, alpha])

    C = np.full(4, pi / 4)

    def outer3(a, b, c):
        """Given 3 1-dimensional vectors a, b, c, the output is of
        shape (len(a), len(b), len(c)) and contains the values

           out[i, j, k] = a[i] * b[j] * c[k]
        """
        return np.multiply.outer(np.multiply.outer(a, b), c)

    r = outer3(rho, sqrt(1 - u ** 2), sqrt(1 - v ** 2))
    s = outer3(rho, sqrt(1 - u ** 2), v)
    t = outer3(rho, u, 4 * [1])

    data = [
        ((A[i] * B[j] * C[k]), np.array([[r[i][j][k], s[i][j][k], t[i][j][k]]]))
        for i in range(4)
        for j in range(4)
        for k in range(4)
    ]
    points, weights = untangle(data)
    d = {"plain": [weights, points[:, 0], points[:, 1], points[:, 2]]}
    weights /= 4 / 3 * math.pi
    return S3Scheme("Stroud S3 7-4", d, 7, _source)


def stroud_14_1():
    # Get the moments corresponding to the Legendre polynomials and the weight function
    # omega(x) = x^2:
    #
    #                                    / 2/3   if k == 0,
    #    int_{-1}^{+1} |x^2| P_k(x) dx ={  8/45  if k == 2,
    #                                    \ 0     otherwise.
    #
    # In this case, the recurrence coefficients can be determined analytically.
    # ```
    # n = 8
    # alpha = np.full(n, fr(0))
    # k = np.arange(n)
    # beta = np.full(n, fr(0))
    # beta[0] = fr(2, 3)
    # # beta[1::2] = fr((k[1::2]+2)**2, ((2*k[1::2]+2)**2 - 1))
    # for k in range(1, n, 2):
    #     beta[k] = fr((k+2)**2, (2*k+2)**2 - 1)
    # # beta[2::2] = fr(k[2::2]**2, ((2*k[2::2]+2)**2 - 1))
    # for k in range(2, n, 2):
    #     beta[k] = fr(k**2, (2*k+2)**2 - 1)
    #
    # # symbolic computation of the points and weights takes 4orever. Keep an eye on
    # # <https://math.stackexchange.com/q/2450401/36678> for a better algorithm to be
    # # implemented in orthopy.
    # flt = np.vectorize(float)
    # alpha = flt(alpha)
    # beta = flt(beta)
    # points, weights = \
    #     orthopy.line.schemes.custom(alpha, beta, mode='numpy')
    #
    # r = points[-4:]
    # A = weights[-4:]
    # ```
    # TODO get symbolic expressions here
    r = np.array(
        [
            3.242534234038097e-01,
            6.133714327005908e-01,
            8.360311073266362e-01,
            9.681602395076261e-01,
        ]
    )
    A = np.array(
        [
            3.284025994586210e-02,
            9.804813271549834e-02,
            1.262636728646019e-01,
            7.618126780737085e-02,
        ]
    )

    spherical_scheme = sphere_stroud.stroud_u3_14_1()
    v = spherical_scheme.points.T
    B = spherical_scheme.weights

    data = [(A[i] * B[j], r[i] * np.array([v[j]])) for i in range(4) for j in range(72)]

    points, weights = untangle(data)
    d = {"plain": [weights, points[:, 0], points[:, 1], points[:, 2]]}
    weights *= 3
    # weights *= 4 * math.pi
    # weights /= 4 / 3 * math.pi
    return S3Scheme("Stroud S3 14-1", d, 14, _source)


register(
    [
        stroud_3_1,
        stroud_5_1,
        stroud_5_2,
        stroud_7_1a,
        stroud_7_1b,
        stroud_7_2,
        stroud_7_3,
        stroud_7_4,
        stroud_14_1,
    ]
)



import math

import numpy as np
import sympy

from ..helpers import article, fsd, untangle, z
from ._helpers import SnScheme

source = article(
    authors=["Preston C. Hammer", "Arthur H. Stroud"],
    title="Numerical Evaluation of Multiple Integrals II",
    journal="Math. Comp.",
    volume="12",
    year="1958",
    pages="272-280",
    url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
)


def hammer_stroud_11n(n, alpha, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    r = sqrt(frac(n + alpha, n + alpha + 2))
    data = [(frac(1, 2 * n), fsd(n, (r, 1)))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Hammer-Stroud 11n", n, weights, points, 3, source)


def hammer_stroud_12n(n, alpha, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt

    r = sqrt(frac(3 * (n + alpha + 2), (n + 2) * (n + alpha + 4)))
    B1 = frac(
        (4 - n) * (n + 2) * (n + alpha) * (n + alpha + 4), 18 * n * (n + alpha + 2) ** 2
    )
    B2 = frac((n + 2) * (n + alpha) * (n + alpha + 4), 36 * n * (n + alpha + 2) ** 2)
    B0 = 1 - 2 * n * B1 - 2 * n * (n - 1) * B2

    data = [(B0, z(n)), (B1, fsd(n, (r, 1))), (B2, fsd(n, (r, 2)))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Hammer-Stroud 12n", n, weights, points, 5, source)



import json
import pathlib

import numpy as np

from ...helpers import article
from .._helpers import T3Scheme, register

source = article(
    authors=["Jan Jaśkowiec", "N. Sukumar"],
    title="High-order cubature rules for tetrahedra",
    journal="Numerical Methods in Engineering",
    volume="121",
    number="11",
    year="2020",
    pages="2418-2436",
    url="https://doi.org/10.1002/nme.6313",
)
# Data from
# https://onlinelibrary.wiley.com/doi/abs/10.1002/nme.6313

this_dir = pathlib.Path(__file__).resolve().parent


def _read(string, tol=1.0e-14):
    filename = f"js{string}.json"
    with open(this_dir / filename) as f:
        data = json.load(f)

    degree = data["degree"]

    points = np.array(data["points"]).T
    points = np.array([points[0], points[1], points[2], 1.0 - np.sum(points, axis=0)])
    weights = np.array(data["weights"])

    d = {"plain": [weights, points[0], points[1], points[2], points[3]]}

    return T3Scheme(f"Jaśkowiec-Sukumar {string}", d, degree, source, tol)


def jaskowiec_sukumar_02():
    return _read("02")


def jaskowiec_sukumar_03():
    return _read("03")


def jaskowiec_sukumar_04():
    return _read("04")


def jaskowiec_sukumar_05():
    return _read("05")


def jaskowiec_sukumar_06():
    return _read("06")


def jaskowiec_sukumar_07():
    return _read("07")


def jaskowiec_sukumar_08():
    return _read("08")


def jaskowiec_sukumar_09():
    return _read("09")


def jaskowiec_sukumar_10():
    return _read("10")


def jaskowiec_sukumar_11():
    return _read("11")


def jaskowiec_sukumar_12():
    return _read("12")


def jaskowiec_sukumar_13():
    return _read("13")


def jaskowiec_sukumar_14():
    return _read("14")


def jaskowiec_sukumar_15():
    return _read("15", 3.687e-11)


def jaskowiec_sukumar_16():
    return _read("16", 2.653e-11)


def jaskowiec_sukumar_17():
    return _read("17", 3.831e-11)


def jaskowiec_sukumar_18():
    return _read("18", 3.470e-11)


def jaskowiec_sukumar_19a():
    return _read("19a", 4.679e-11)


def jaskowiec_sukumar_19b():
    return _read("19b", 4.580e-11)


def jaskowiec_sukumar_20():
    return _read("20", 2.315e-11)


register(
    [
        jaskowiec_sukumar_02,
        jaskowiec_sukumar_03,
        jaskowiec_sukumar_04,
        jaskowiec_sukumar_05,
        jaskowiec_sukumar_06,
        jaskowiec_sukumar_07,
        jaskowiec_sukumar_08,
        jaskowiec_sukumar_09,
        jaskowiec_sukumar_10,
        jaskowiec_sukumar_11,
        jaskowiec_sukumar_12,
        jaskowiec_sukumar_13,
        jaskowiec_sukumar_14,
        jaskowiec_sukumar_15,
        jaskowiec_sukumar_16,
        jaskowiec_sukumar_17,
        jaskowiec_sukumar_18,
        jaskowiec_sukumar_19a,
        jaskowiec_sukumar_19b,
        jaskowiec_sukumar_20,
    ]
)



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["J. Albrecht", "L. Collatz"],
    title="Zur numerischen Auswertung mehrdimensionaler Integrale",
    journal="ZAMM",
    volume="38",
    number="1-2",
    year="1958",
    pages="1–15",
    url="https://doi.org/10.1002/zamm.19580380102",
)


def albrecht_collatz_1():
    d = {
        "zero2": [[frac(5, 12)]],
        "d4_a0": [[frac(1, 8)], [1]],
        "d4_aa": [[frac(1, 48)], [1]],
    }
    return C2Scheme("Albrecht-Collatz 1", d, 3, source, 4.442e-16)


def albrecht_collatz_2():
    r = sqrt(frac(3, 5))
    s = sqrt(frac(1, 3))
    t = sqrt(frac(14, 15))
    d = {
        "zero2": [[frac(2, 7)]],
        "c2": [[frac(5, 63)], [0], [t]],
        "sxy": [[frac(5, 36)], [r], [s]],
    }
    return C2Scheme("Albrecht-Collatz 2", d, 5, source, 4.627e-16)


def albrecht_collatz_3():
    r = sqrt(frac(7, 15))
    s, t = (sqrt((7 + i * sqrt(24)) / 15) for i in [+1, -1])
    d = {
        "zero2": [[frac(2, 7)]],
        "c2": [[frac(25, 168), frac(5, 48), frac(5, 48)], [r, s, t], [r, -t, -s]],
    }
    return C2Scheme("Albrecht-Collatz 3", d, 5, source, 4.442e-16)


def albrecht_collatz_4():
    d = {
        "zero2": [[frac(2, 45)]],
        "d4_a0": [[frac(2, 45)], [1]],
        "d4_aa": [[frac(1, 60), frac(8, 45)], [1, frac(1, 2)]],
    }
    return C2Scheme("Albrecht-Collatz 4", d, 5, source, 8.883e-16)


register(
    [albrecht_collatz_1, albrecht_collatz_2, albrecht_collatz_3, albrecht_collatz_4]
)



from sympy import Rational as frac

from ..helpers import article
from ._helpers import T2Scheme, register

source = article(
    authors=["J. Albrecht", "L. Collatz"],
    title="Zur numerischen Auswertung mehrdimensionaler Integrale",
    journal="ZAMM",
    volume="38",
    number="1-2",
    year="1958",
    pages="1–15",
    url="https://doi.org/10.1002/zamm.19580380102",
)


def albrecht_collatz():
    d = {"d3_aa": [[frac(1, 30), frac(9, 30)], [frac(1, 2), frac(1, 6)]]}
    return T2Scheme("Albrecht-Collatz", d, 3, source, tol=2.776e-16)


register([albrecht_collatz])



import numpy as np

from ..helpers import article
from ._helpers import S2Scheme, register

_source = article(
    authors=["W.H. Peirce"],
    title="Numerical integration over the planar annulus",
    journal="J. Soc. Indust. Appl. Math.",
    volume="5",
    number="2",
    month="jun",
    year="1957",
    url="https://www.jstor.org/stable/2098722",
)


# TODO symbolic
def peirce_1957(m):
    k = 4 * m + 3
    theta = 2 * np.pi * np.arange(1, k + 2) / (k + 1)
    p, w = np.polynomial.legendre.leggauss(m + 1)
    # scale points to [r0, r1] (where r0 = 0, r1 = 1 for now)
    p = np.sqrt(0.5 * (p + 1.0))
    p_theta = np.dstack(np.meshgrid(p, theta)).reshape(-1, 2).T
    points = np.array(
        [p_theta[0] * np.cos(p_theta[1]), p_theta[0] * np.sin(p_theta[1])]
    )
    # When integrating between 0 and 1, the weights are exactly the Gauss-Legendre
    # weights, scaled according to the disk area.
    weights = np.tile(0.5 / (k + 1) * w, k + 1)
    d = {"plain": [weights, points[0], points[1]]}
    return S2Scheme("Peirce 1957", d, k, _source)


register([peirce_1957])



import pathlib

from ...helpers import techreport
from .._helpers import _read, register

_source = techreport(
    authors=["R. Cools", "A. Haegemans"],
    title="Construction of fully symmetric cubature formulae of degree 4k-3 for fully symmetric planar regions",
    year="1985",
    institution="Dept. of Computer Science, KU Leuven",
    number="Report TW 71",
    url="https://lirias.kuleuven.be/bitstream/123456789/131870/1/TW71.pdf",
)

this_dir = pathlib.Path(__file__).resolve().parent


def cools_haegemans_9_1():
    return _read(this_dir / "cools_haegemans_09_1.json", _source)


def cools_haegemans_9_2():
    return _read(this_dir / "cools_haegemans_09_2.json", _source)


def cools_haegemans_13_1():
    return _read(this_dir / "cools_haegemans_13_1.json", _source)


register([cools_haegemans_9_1, cools_haegemans_9_2, cools_haegemans_13_1])



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, untangle, z
from ._helpers import CnScheme

_source = article(
    authors=["L.N. Dobrodeev"],
    title="Cubature formulas of the seventh order of accuracy for a hypersphere and a hypercube",
    journal="USSR Computational Mathematics and Mathematical Physics",
    volume="10",
    number="1",
    year="1970",
    pages="252–253",
    url="https://doi.org/10.1016/0041-5553%2870%2990084-4",
)


def dobrodeev_1970(n):
    assert n >= 3, f"Only works for n >= 3, not n = {n}"

    A = frac(1, 8)
    B = frac(19 - 5 * n, 20)
    alpha = 35 * n * (5 * n - 33)
    C = frac((alpha + 2114) ** 3, 700 * (alpha + 1790) * (alpha + 2600))
    D = frac(729, 1750) * frac(alpha + 2114, alpha + 2600)
    E = frac(n * (n - 1) * (n - frac(47, 10)), 3) - 2 * n * (C + D) + frac(729, 125)

    a = sqrt(frac(3, 5))
    b = a
    c = sqrt(frac(3, 5) * frac(alpha + 1790, alpha + 2114))
    d = 1
    data = [
        (A, fsd(n, (a, 3))),
        (B, fsd(n, (b, 2))),
        (C, fsd(n, (c, 1))),
        (D, fsd(n, (d, 1))),
        (E, z(n)),
    ]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    weights *= frac(125, 729)

    # d = {
    #     "0": [[w0]],
    #     "a0": [[w1], [lmbdas[0]]],
    #     "aa0": [[w1], [lmbdas[0]]],
    #     "aaa0": [[w1], [lmbdas[0]]],
    #     "a": [[w], [sqrt(delta2)]]
    # }
    return CnScheme("Dobrodeev 1970", n, weights, points, 7, _source, 8.100e-13)



import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import book, fsd, pm, untangle
from ._helpers import EnrScheme
from ._stroud_secrest import stroud_secrest_2 as stroud_enr_3_1
from ._stroud_secrest import stroud_secrest_3 as stroud_enr_3_2
from ._stroud_secrest import stroud_secrest_4 as stroud_enr_5_1

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)

# ERR
# TODO find mistake
# def _gen5_2(n):
#     assert n != 3
#
#     r2 = -(n+1)*(n+3) + (n+3)*sqrt((n+1)*(2*n+3))
#     s2 = n*(n+1)*(n+3) - 2*(n+3)*sqrt((n+1)*(2*n+3)) / (n-3) / (n+2)
#     A = (n+1) * (n+4) / r2**2
#     B = (n+1) * (n+3) / 2**n / s2**2
#     r = sqrt(r2)
#     s = sqrt(s2)
#
#     data = [
#         (A, fsd(n, (r, 1))),
#         (B, pm(n, s)),
#         ]
#     return 5, data


def stroud_enr_5_3(n):
    """Spherical product Lobatto formula."""
    data = []
    s = sqrt(n + 3)
    for k in range(1, n + 1):
        rk = sqrt((k + 2) * (n + 3))
        Bk = frac(2 ** (k - n) * (n + 1), (k + 1) * (k + 2) * (n + 3))
        arr = (k - 1) * [0] + [rk] + (n - k) * [s]
        data += [(Bk, pm(arr))]
    B0 = 1 - sum(item[0] * len(item[1]) for item in data)
    data += [(B0, np.full((1, n), 0))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return EnrScheme("Stroud Enr 5-3", n, weights, points, 5, source)


def stroud_enr_5_4(n):
    r = sqrt(((n + 2) * (n + 3) + (n - 1) * (n + 3) * sqrt(2 * (n + 2))) / n)
    s = sqrt(((n + 2) * (n + 3) - (n + 3) * sqrt(2 * (n + 2))) / n)
    A = frac(4 * n + 6, (n + 2) * (n + 3))
    B = frac(n + 1, (n + 2) * (n + 3) * 2 ** n)
    data = [(A, np.full((1, n), 0)), (B, fsd(n, (r, 1), (s, n - 1)))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return EnrScheme("Stroud Enr 5-4", n, weights, points, 5, source, 1.684e-11)


# math domain error
# def _gen5_5(n):
#     r = sqrt((n*(n+1) - sqrt((n+1)*(4*n+6)) + (n-1)*(n+1)*sqrt(4*n+6)) / n)
#     s = sqrt((n*(n+1) - sqrt((n+1)*(4*n+6)) - (n+1)*sqrt(4*n+6)) / n)
#     t = n + 1 + sqrt((n+1)*(4*n+6))
#     A = 1 / (n+1) / 2**n
#     data = [
#         (A, fsd(n, (r, 1), (s, n-1))),
#         (A, pm(n, t)),
#         ]
#     return 5, data


# TODO find out what's wrong
# def _gen7_1(n):
#     assert 3 <= n <= 7
#
#     alpha = sqrt(3*(n+3)*(2*n+7)*(8-n))
#
#     # One could change some signs here and the formula would still work, but
#     # weights and points would be complex-valued.
#     r2 = (n+5) * (3*(n+3)*(8-n) - (n-2) * alpha) / (-2*n**2 + 6*n + 11)
#     s2 = (n+5) * (3*n*(n+3) - 2 * alpha) / (3*n**2 + 5*n - 56)
#     t2 = (n+5) * (6*(n+3) + 2 * alpha) / (2*n - 5)
#
#     B = (8-n) * (n+1) * (n+3) * (n+5) / r2**3
#     C = (n+1) * (n+3) * (n+5) / 2**n / s2**3
#     D = (n+1) * (n+3) * (n+5) / 2 / t2**3
#     A = 1.0 - 2*n*B - 2**n*C - 2*n*(n-1)*D
#
#     r = sqrt(r2)
#     s = sqrt(s2)
#     t = sqrt(t2)
#
#     data = [
#         (A, np.full((1, n), 0.0)),
#         (B, fsd(n, (r, 1))),
#         (C, pm(n, s)),
#         (D, fsd(n, (t, 2))),
#         ]
#     return 7, data


__all__ = [
    "stroud_enr_3_1",
    "stroud_enr_3_2",
    "stroud_enr_5_1",
    "stroud_enr_5_3",
    "stroud_enr_5_4",
]



import numpy as np
import sympy

from ..helpers import article
from ._helpers import S2Scheme, register

_source = article(
    authors=["J. Radon"],
    title="Zur mechanischen Kubatur",
    journal="Monatshefte für Mathematik",
    year="1948",
    volume="52",
    pages="286-300",
    issn="0026-9255",
    issne="1436-5081/e",
    url="https://eudml.org/doc/176796",
)

frac = sympy.Rational
sqrt = np.vectorize(sympy.sqrt)


def radon(alpha=0):
    r = sqrt(frac(alpha + 4, alpha + 6))
    s = sqrt(frac(alpha + 4, 4 * (alpha + 6)))
    t = sqrt(frac(3 * (alpha + 4), 4 * (alpha + 6)))

    A = frac(4, (alpha + 4) ** 2)
    B = frac((alpha + 2) * (alpha + 6), 6 * (alpha + 4) ** 2)

    d = {
        "zero2": [[A]],
        "c2_a0": [[B], [r]],
        # ERR Stroud is missing +- in front of t.
        "sxy": [[B], [s], [t]],
    }
    return S2Scheme(f"Radon({alpha})", d, 5, _source)


register([radon])



import ndim
import numpy as np

from ..helpers import QuadratureScheme


class UnScheme(QuadratureScheme):
    def __init__(self, name, dim, weights, points, degree, source, tol=1.0e-14):
        super().__init__(name, weights, points, degree, source, tol)
        self.domain = f"Un (n={dim})"
        self.dim = dim

    def integrate(self, f, center, radius, dot=np.dot):
        center = np.array(center)
        rr = np.multiply.outer(radius, self.points)
        rr = np.swapaxes(rr, 0, -2)
        ff = np.array(f((rr + center).T))
        ref_vol = ndim.nsphere.volume(self.dim, r=radius)
        return ref_vol * dot(ff, self.weights)



from ._hammer_marlowe_stroud import hammer_marlowe_stroud_1 as hammer_stroud_2
from ._hammer_marlowe_stroud import hammer_marlowe_stroud_3 as hammer_stroud_3

__all__ = ["hammer_stroud_2", "hammer_stroud_3"]

# source = article(
#     authors=["Preston C. Hammer", "Arthur H. Stroud"],
#     title="Numerical Evaluation of Multiple Integrals II",
#     journal="Math. Comp.",
#     volume="12",
#     year="1958",
#     pages="272-280",
#     url="https://doi.org/10.1090/S0025-5718-1958-0102176-6",
# )



import itertools

import numpy as np


def _zero(data, dim):
    return np.zeros((dim, 1))


def _d4_ab(data, dim):
    """dihedral symmetry d4."""
    s, t = np.array(data)
    points = np.array(
        [[+s, +t], [-s, +t], [+s, -t], [-s, -t], [+t, +s], [-t, +s], [+t, -s], [-t, -s]]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _d4_aa(data, dim):
    a = np.asarray(data)
    points = np.array([[+a, +a], [-a, +a], [+a, -a], [-a, -a]])
    points = np.moveaxis(points, 0, 1)
    return points


def _d4_a0(data, dim):
    a = np.asarray(data)
    zero = np.zeros_like(a)
    points = np.array([[+a, zero], [-a, zero], [zero, +a], [zero, -a]])
    points = np.moveaxis(points, 0, 1)
    return points


def _c4(data, dim):
    a, b = data
    points = np.array([[+a, +b], [-a, -b], [-b, +a], [+b, -a]])
    points = np.moveaxis(points, 0, 1)
    return points


def _sxy(data, dim):
    x, y = data
    points = np.array([[+x, +y], [+x, -y], [-x, +y], [-x, -y]])
    points = np.moveaxis(points, 0, 1)
    return points


def _sx(data, dim):
    x, y = data
    points = np.array([[+x, y], [-x, y]])
    points = np.moveaxis(points, 0, 1)
    return points


def _c2(data, dim):
    x, y = data
    points = np.array([[+x, +y], [-x, -y]])
    points = np.moveaxis(points, 0, 1)
    return points


def _c2_a0(r, dim):
    zero = np.zeros_like(r)
    points = np.array([[+r, zero], [-r, zero]])
    points = np.moveaxis(points, 0, 1)
    return points


def _c2_0a(data, dim):
    a = np.asarray(data)
    zero = np.zeros_like(a)
    points = np.array([[zero, +a], [zero, -a]])
    points = np.moveaxis(points, 0, 1)
    return points


def _centroid(data, dim):
    return np.full((3, 1), 1 / 3)


def _vertex(data, dim):
    return np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])


def _d3_aa(a, dim):
    a = np.array(a)
    b = 1 - 2 * a
    return np.array([[a, a, b], [a, b, a], [b, a, a]])


def _d3_ab(data, dim):
    a, b = np.asarray(data)
    c = 1 - a - b
    points = np.array(
        [[a, b, c], [c, a, b], [b, c, a], [b, a, c], [c, b, a], [a, c, b]]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _c3_ab(data, dim):
    a, b = data
    c = 1 - a - b
    points = np.array([[a, b, c], [c, a, b], [b, c, a]])
    points = np.moveaxis(points, 0, 1)
    return points


def _swap_ab(data, dim):
    a, b = data
    c = 1 - a - b
    points = np.array([[a, b, c], [b, a, c]])
    points = np.moveaxis(points, 0, 1)
    return points


def _s2_static(a, dim):
    a = np.asarray(a)
    b = 1 - 2 * a
    points = np.array([[a, a, b]])
    points = np.moveaxis(points, 0, 1)
    return points


def _d(n, offset, r):
    import sympy

    cos = np.vectorize(sympy.cos)
    sin = np.vectorize(sympy.sin)

    alpha = (2 * np.arange(n) + offset) * sympy.pi / n
    cs = np.array([cos(alpha), sin(alpha)])
    points = np.multiply.outer(cs, r)
    return points


def _symm_r00(r, dim):
    zero = np.zeros_like(r)
    points = np.array(
        [
            [+r, zero, zero],
            [-r, zero, zero],
            [zero, +r, zero],
            [zero, -r, zero],
            [zero, zero, +r],
            [zero, zero, -r],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _symm_rr0(a, dim):
    z = np.zeros_like(a)
    points = np.array(
        [
            [+a, +a, z],
            [+a, z, +a],
            [z, +a, +a],
            [+a, -a, z],
            [+a, z, -a],
            [z, +a, -a],
            [-a, +a, z],
            [-a, z, +a],
            [z, -a, +a],
            [-a, -a, z],
            [-a, z, -a],
            [z, -a, -a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _symm_rs0_roll(data, dim):
    r, s = data
    z = np.zeros_like(r)
    points = np.array(
        [
            [+r, +s, z],
            [+r, -s, z],
            [-r, +s, z],
            [-r, -s, z],
            [z, +r, +s],
            [z, +r, -s],
            [z, -r, +s],
            [z, -r, -s],
            [+s, z, +r],
            [+s, z, -r],
            [-s, z, +r],
            [-s, z, -r],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _symm_rrr(a, dim):
    points = np.array(
        [
            [+a, +a, +a],
            [-a, +a, +a],
            [+a, -a, +a],
            [-a, -a, +a],
            [+a, +a, -a],
            [-a, +a, -a],
            [+a, -a, -a],
            [-a, -a, -a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _symm_rrs(data, dim):
    a, b = data
    points = np.array(
        [
            [+a, +a, +b],
            [+a, +b, +a],
            [+b, +a, +a],
            [+a, -a, +b],
            [+a, +b, -a],
            [+b, +a, -a],
            [-a, +a, +b],
            [-a, +b, +a],
            [+b, -a, +a],
            [-a, -a, +b],
            [-a, +b, -a],
            [+b, -a, -a],
            [+a, +a, -b],
            [+a, -b, +a],
            [-b, +a, +a],
            [+a, -a, -b],
            [+a, -b, -a],
            [-b, +a, -a],
            [-a, +a, -b],
            [-a, -b, +a],
            [-b, -a, +a],
            [-a, -a, -b],
            [-a, -b, -a],
            [-b, -a, -a],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _symm_rss_pm(data, dim):
    r, s = data
    points = np.array(
        [
            [+r, +s, +s],
            [+s, +r, +s],
            [+s, +s, +r],
            [-r, -s, -s],
            [-s, -r, -s],
            [-s, -s, -r],
        ]
    )
    points = np.moveaxis(points, 0, 1)
    return points


def _a(a, dim):
    assert a.shape == (1, 1)
    a = a[0][0]
    possible_vals = dim * [[+a, -a]]
    out = np.array(list(itertools.product(*possible_vals))).T
    return out


def _a0(a, dim):
    assert a.shape == (1, 1)
    a = a[0][0]
    entries = []
    for i in range(dim):
        entries += [
            i * [0] + [+a] + (dim - i - 1) * [0],
            i * [0] + [-a] + (dim - i - 1) * [0],
        ]
    return np.array(entries).T


def expand_symmetries_points_only(data, dim):
    points = []
    counts = []

    for key, points_raw in data.items():
        fun = {
            "zero2": _zero,
            "zero3": _zero,
            "0": _zero,
            #
            "a": _a,
            "a0": _a0,
            #
            "d4_aa": _d4_aa,
            "d4_a0": _d4_a0,
            "c4": _c4,
            "d4_ab": _d4_ab,
            "c2_a0": _c2_a0,
            "c2_0a": _c2_0a,
            "sxy": _sxy,
            "c2": _c2,
            "sx": _sx,
            #
            "centroid": _centroid,
            "vertex": _vertex,
            "d3_ab": _d3_ab,
            "d3_aa": _d3_aa,
            "c3_ab": _c3_ab,
            "swap_ab": _swap_ab,
            "s2_static": _s2_static,
            #
            "d4.0": lambda r, dim: _d(4, 0, r),
            "d4.1": lambda r, dim: _d(4, 1, r),
            "d5.0": lambda r, dim: _d(5, 0, r),
            "d6.0": lambda r, dim: _d(6, 0, r),
            "d6.1": lambda r, dim: _d(6, 1, r),
            "d8.0": lambda r, dim: _d(8, 0, r),
            "d8.1": lambda r, dim: _d(8, 1, r),
            "d10.0": lambda r, dim: _d(10, 0, r),
            "d10.1": lambda r, dim: _d(10, 1, r),
            #
            "symm_r00": _symm_r00,
            "symm_rr0": _symm_rr0,
            "symm_rs0_roll": _symm_rs0_roll,
            "symm_rrr": _symm_rrr,
            "symm_rrs": _symm_rrs,
            "symm_rss_pm": _symm_rss_pm,
            #
            "plain": lambda vals, dim: vals.reshape(vals.shape[0], 1, -1),
        }[key]
        pts = fun(np.asarray(points_raw), dim)

        counts.append(pts.shape[1])
        pts = pts.reshape(pts.shape[0], -1)
        points.append(pts)

    points = np.ascontiguousarray(np.concatenate(points, axis=1))
    return points, counts


def expand_symmetries(data, dim):
    # separate points and weights
    points_raw = {}
    weights_raw = []
    for key, values in data.items():
        weights_raw.append(values[0])
        points_raw[key] = values[1:]

    points, counts = expand_symmetries_points_only(points_raw, dim)
    weights = np.concatenate(
        [np.tile(values, count) for count, values in zip(counts, weights_raw)]
    )
    return points, weights



from sympy import Rational as frac
from sympy import sqrt

from ..cn import ewing
from ..helpers import book
from ._albrecht_collatz import albrecht_collatz as stroud_c3_3_6
from ._hammer_stroud import hammer_stroud_2_3 as stroud_c3_5_2
from ._hammer_stroud import hammer_stroud_5_3a as stroud_c3_7_1a
from ._hammer_stroud import hammer_stroud_5_3b as stroud_c3_7_1b
from ._hammer_wymore import hammer_wymore as stroud_c3_7_2
from ._helpers import C3Scheme, register
from ._mustard_lyness_blatt import mustard_lyness_blatt_1 as stroud_c3_3_4
from ._mustard_lyness_blatt import mustard_lyness_blatt_2 as stroud_c3_3_5
from ._mustard_lyness_blatt import mustard_lyness_blatt_3 as stroud_c3_3_7
from ._mustard_lyness_blatt import mustard_lyness_blatt_4 as stroud_c3_5_4
from ._mustard_lyness_blatt import mustard_lyness_blatt_5 as stroud_c3_5_5
from ._mustard_lyness_blatt import mustard_lyness_blatt_6 as stroud_c3_5_6
from ._mustard_lyness_blatt import mustard_lyness_blatt_7 as stroud_c3_5_7
from ._sadowsky import sadowsky as stroud_c3_5_8
from ._sarma_stroud import sarma_stroud as stroud_c3_7_3
from ._stroud_1967 import stroud_1967 as stroud_c3_5_1
from ._tyler import tyler_1 as stroud_c3_3_1
from ._tyler import tyler_2 as stroud_c3_5_3

_source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_c3_3_2():
    # Product Gauss scheme
    d = {"symm_rrr": [[frac(1, 8)], [sqrt(frac(1, 3))]]}
    return C3Scheme("Stroud C3 3-2", d, 3, _source)


def stroud_c3_3_3():
    return ewing(3)


register(
    [
        stroud_c3_3_1,
        stroud_c3_3_2,
        stroud_c3_3_3,
        stroud_c3_3_4,
        stroud_c3_3_5,
        stroud_c3_3_6,
        stroud_c3_3_7,
        stroud_c3_5_1,
        stroud_c3_5_2,
        stroud_c3_5_3,
        stroud_c3_5_4,
        stroud_c3_5_5,
        stroud_c3_5_6,
        stroud_c3_5_7,
        stroud_c3_5_8,
        stroud_c3_7_1a,
        stroud_c3_7_1b,
        stroud_c3_7_2,
        stroud_c3_7_3,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Linbo Zhang", "Tao Cui", "Hui Liu"],
    title="A set of symmetric quadrature rules on triangles and tetrahedra",
    journal="Journal of Computational Mathematics",
    volume="27",
    number="1",
    month="jan",
    year="2009",
    pages="89-96",
    url="https://www.jstor.org/stable/43693493",
)

this_dir = pathlib.Path(__file__).resolve().parent


def zhang_cui_liu_1():
    return _read(this_dir / "zhang_cui_liu_1.json", source)


def zhang_cui_liu_2():
    return _read(this_dir / "zhang_cui_liu_2.json", source)


def zhang_cui_liu_3():
    return _read(this_dir / "zhang_cui_liu_3.json", source)


register([zhang_cui_liu_1, zhang_cui_liu_2, zhang_cui_liu_3])



import itertools
from typing import Callable

import numpy as np

from .._exception import QuadpyError
from ..helpers import QuadratureScheme, n_outer


class CnScheme(QuadratureScheme):
    def __init__(
        self,
        name: str,
        dim: int,
        weights,
        points,
        degree: int,
        source,
        tol=1.0e-14,
        comments=None,
    ):
        self.domain = f"Cn (n={dim})"
        self.dim = dim
        assert points.shape[0] == dim, f"points.shape == {points.shape}, dim = {dim}"
        super().__init__(name, weights, points, degree, source, tol, comments)

    def integrate(self, f: Callable, ncube, dot=np.dot):
        ncube = np.asarray(ncube)
        if (
            ncube.shape[: self.dim] != tuple(self.dim * [2])
            or ncube.shape[-1] < self.dim
        ):
            expected = ", ".join(str(i) for i in self.dim * [2])
            string = ", ".join(str(val) for val in ncube.shape)
            raise QuadpyError(
                f"Wrong domain shape. Expected ({expected}, ..., n >= {self.dim}), "
                f"got ({string})."
            )
        x = transform(self.points, ncube).T
        detJ = get_detJ(self.points, ncube)

        fx = np.asarray(f(x))
        if fx.shape[-len(x.shape[1:]) :] != x.shape[1:]:
            string = ", ".join(str(val) for val in x.shape[1:])
            raise QuadpyError(
                f"Wrong return value shape {fx.shape}. Expected (..., {string})."
            )

        ref_vol = 2 ** np.prod(self.dim)
        return ref_vol * dot(fx * abs(detJ), self.weights)

    def points_inside(self):
        """Are all points strictly inside the domain?"""
        return np.all((-1 < self.points) & (self.points < 1))

    def points_inside_or_boundary(self):
        """Are all points inside the domain or on its boundary?"""
        return np.all((-1 <= self.points) & (self.points <= 1))


def transform(xi, cube):
    """Transform the points `xi` from the reference cube to `cube`."""
    # For d==2, the result used to be computed with
    #
    # out = (
    #     + outer(0.25*(1.0-xi[0])*(1.0-xi[1]), cube[0, 0])
    #     + outer(0.25*(1.0+xi[0])*(1.0-xi[1]), cube[1, 0])
    #     + outer(0.25*(1.0-xi[0])*(1.0+xi[1]), cube[0, 1])
    #     + outer(0.25*(1.0+xi[0])*(1.0+xi[1]), cube[1, 1])
    #     )
    #
    # This array of multiplications and additions is reminiscent of dot(), and
    # indeed tensordot() can handle the situation. We just need to compute the
    # `1+-xi` products and align them with `cube`.
    one_mp_xi = np.stack([0.5 * (1.0 - xi), 0.5 * (1.0 + xi)], axis=1)
    a = n_outer(one_mp_xi)

    # <https://stackoverflow.com/q/45372098/353337>
    d = xi.shape[0]
    return np.tensordot(a, cube, axes=(range(d), range(d)))


def get_detJ(xi, cube):
    """Get the determinant of the transformation matrix."""
    # For d==2, the result can be computed with
    # ```
    # J0 = (
    #     - np.multiply.outer(0.25*(1-xi[1]), quad[0, 0])
    #     + np.multiply.outer(0.25*(1-xi[1]), quad[1, 0])
    #     - np.multiply.outer(0.25*(1+xi[1]), quad[0, 1])
    #     + np.multiply.outer(0.25*(1+xi[1]), quad[1, 1])
    #     ).T
    # J1 = (
    #     - np.multiply.outer(0.25*(1-xi[0]), quad[0, 0])
    #     - np.multiply.outer(0.25*(1+xi[0]), quad[1, 0])
    #     + np.multiply.outer(0.25*(1-xi[0]), quad[0, 1])
    #     + np.multiply.outer(0.25*(1+xi[0]), quad[1, 1])
    #     ).T
    # out = J0[0]*J1[1] - J1[0]*J0[1]
    # ```
    # Like transform(), simplify here and form the determinant explicitly.
    d = xi.shape[0]

    one_mp_xi = np.stack([0.5 * (1.0 - xi), 0.5 * (1.0 + xi)], axis=1)

    # Build the Jacobi matrix row by row.
    J = []
    for k in range(d):
        a = one_mp_xi.copy()
        a[k, 0, :] = -0.5
        a[k, 1, :] = +0.5
        a0 = n_outer(a)
        J.append(np.tensordot(a0, cube, axes=(range(d), range(d))).T)

    # `det` needs the square at the end. Fortran...
    # For d==2 or d==3, we could avoid this copy and compute the determinant
    # with their elementary formulas, i.e.,
    #
    #     + J[0][0]*J[1][1] - J[1][0]*J[0][1];
    #
    #     + J0[0]*J1[1]*J2[2] + J1[0]*J2[1]*J0[2] + J2[0]*J0[1]*J1[2]
    #     - J0[2]*J1[1]*J2[0] - J1[2]*J2[1]*J0[0] - J2[2]*J0[1]*J1[0].
    #
    J = np.array(J)
    J = np.moveaxis(J, (0, 1), (-2, -1))
    out = np.linalg.det(J)
    return out


def integrate_monomial_over_ncube(ncube_limits, exp):
    return np.prod(
        [
            (a[1] ** (k + 1) - a[0] ** (k + 1)) / (k + 1)
            for a, k in zip(ncube_limits, exp)
        ]
    )


def ncube_points(*xyz):
    """Given the end points of an n-cube aligned with the coordinate axes, this returns
    the corner points of the cube in the correct data structure.
    """
    return np.moveaxis(np.array(np.meshgrid(*xyz, indexing="ij")), 0, -1)


def _fs11(n, r, s):
    """Get all permutations of [+-r, +-s, ..., +-s] of length n.
    len(out) == n * 2**n.
    """
    # <https://stackoverflow.com/a/45321972/353337>
    k1 = 1
    k2 = n - 1
    idx = itertools.combinations(range(k1 + k2), k1)
    vs = ((s if j not in i else r for j in range(k1 + k2)) for i in idx)
    return np.array(
        list(
            itertools.chain.from_iterable(
                itertools.product(*((+vij, -vij) for vij in vi)) for vi in vs
            )
        )
    )


def _s(n, a, b):
    """Get all permutations of [a, b, ..., b] of length n. len(out) == n."""
    out = np.full((n, n), b)
    np.fill_diagonal(out, a)
    return out


def _s2(n, a):
    """Get all permutations of [a, a, 0, 0, ..., 0] of length n.
    len(out) == (n-1)*n / 2.
    """
    return _s11(n, a, a)


def _s11(n, a, b):
    """Get all permutations of [a, b, 0, 0, ..., 0] of length n.
    len(out) == (n-1)*n.
    """
    s = [a, b] + (n - 2) * [0]
    # First permutations, then set can be really inefficient if items are
    # repeated. Check out <https://stackoverflow.com/q/6284396/353337> for
    # improvements.
    return np.array(list(set(itertools.permutations(s, n))))



import math
import sys
from collections import namedtuple

import numpy as np
import sympy

article = namedtuple(
    "Article",
    [
        "authors",
        "title",
        "journal",
        "volume",
        "number",
        "year",
        "month",
        "pages",
        "url",
        "issn",
        "issne",
        "note",
    ],
    # defaults=(None,) * 8
)
article.__new__.__defaults__ = (None,) * len(article._fields)

book = namedtuple(
    "Book", ["authors", "title", "publisher", "isbn", "year", "url", "note"]
)
book.__new__.__defaults__ = (None,) * len(book._fields)

techreport = namedtuple(
    "Techreport",
    ["authors", "title", "year", "month", "institution", "number", "url", "note"],
)
techreport.__new__.__defaults__ = (None,) * len(techreport._fields)

phdthesis = namedtuple(
    "Phdthesis", ["authors", "title", "year", "school", "url", "note"]
)
phdthesis.__new__.__defaults__ = (None,) * len(phdthesis._fields)

online = namedtuple("Online", ["authors", "title", "year", "url", "note"])
online.__new__.__defaults__ = (None,) * len(online._fields)


def untangle(data):
    weights, points = zip(*data)
    return (
        np.concatenate(points),
        np.repeat(weights, [len(grp) for grp in points]),
    )


def n_outer(a):
    """Given a list (tuple, array) of arrays, this method computes their outer
    product. If the dimension of the input arrays is larger than one, the
    product is formed across the first dimension; all other dimensions must
    coincide in size.

    Examples:
    n_outer([np.ones(4), np.ones(5)]).shape == (4, 5)
    n_outer([np.ones(4), np.ones(5), np.ones(6)]).shape == (4, 5, 6)
    n_outer([np.ones(4, 3, 7), np.ones(5, 3, 7)]).shape == (4, 5, 3, 7)
    """
    # <https://stackoverflow.com/a/45376730/353337>
    d = len(a)

    # If the elements are more than one-dimensional, assert that the extra
    # dimensions are all equal.
    s0 = a[0].shape
    for arr in a:
        assert s0[1:] == arr.shape[1:]

    out = a[0]
    for k in range(1, d):
        # Basically outer products. Checkout `np.outer`'s implementation for
        # comparison.
        out = np.multiply(
            # Insert a newaxis after k `:`
            out[(slice(None),) * k + (np.newaxis,)],
            # Insert a newaxis at the beginning
            a[k][np.newaxis],
        )
    return out


def compute_dobrodeev(n, I0, I2, I22, I4, pm_type, i, j, k, symbolic=False):
    """Compute some helper quantities used in

    L.N. Dobrodeev,
    Cubature rules with equal coefficients for integrating functions with
    respect to symmetric domains,
    USSR Computational Mathematics and Mathematical Physics,
    Volume 18, Issue 4, 1978, Pages 27-34,
    <https://doi.org/10.1016/0041-5553%2878%2990064-2>.
    """
    t = 1 if pm_type == "I" else -1

    fact = sympy.factorial if symbolic else math.factorial
    sqrt = sympy.sqrt if symbolic else np.sqrt

    L = comb(n, i) * 2 ** i
    M = fact(n) // (fact(j) * fact(k) * fact(n - j - k)) * 2 ** (j + k)
    N = L + M
    F = I22 / I0 - I2 ** 2 / I0 ** 2 + (I4 / I0 - I22 / I0) / n
    R = (
        -(j + k - i) / i * I2 ** 2 / I0 ** 2
        + (j + k - 1) / n * I4 / I0
        - (n - 1) / n * I22 / I0
    )
    H = (
        1
        / i
        * (
            (j + k - i) * I2 ** 2 / I0 ** 2
            + (j + k) / n * ((i - 1) * I4 / I0 - (n - 1) * I22 / I0)
        )
    )
    Q = L / M * R + H - t * 2 * I2 / I0 * (j + k - i) / i * sqrt(L / M * F)

    G = 1 / N
    a = sqrt(n / i * (I2 / I0 + t * sqrt(M / L * F)))
    b = sqrt(n / (j + k) * (I2 / I0 - t * sqrt(L / M * F) + t * sqrt(k / j * Q)))
    c = sqrt(n / (j + k) * (I2 / I0 - t * sqrt(L / M * F) - t * sqrt(j / k * Q)))
    return G, a, b, c


def get_nsimplex_points(n, sqrt, frac):
    # vertices of the n-simplex
    points = []
    for r in range(n + 1):
        point = [-sqrt(frac(n + 1, n * (n - i + 1) * (n - i))) for i in range(r)]
        if r < n:
            point += [sqrt(frac((n + 1) * (n - r), n * (n - r + 1)))]
            point += [0] * (n - r - 1)
        points.append(point)
    return np.array(points)


def prod(lst):
    if sys.version < "3.8":
        import operator
        from functools import reduce

        return reduce(operator.mul, lst, 1)
    return math.prod(lst)


def comb(a, b):
    if sys.version < "3.8":
        try:
            binom = math.factorial(a) // math.factorial(b) // math.factorial(a - b)
        except ValueError:
            binom = 0
        return binom
    return math.comb(a, b)


def gamma_n_2(n, symbolic):
    # gamma(n / 2)
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    pi = sympy.pi if symbolic else math.pi

    if n % 2 == 0:
        return math.factorial(n // 2 - 1)

    n2 = n // 2
    return frac(math.factorial(2 * n2), 4 ** n2 * math.factorial(n2)) * sqrt(pi)



from sympy import Rational as frac

from ..helpers import article
from ._helpers import C2Scheme, register

source = article(
    authors=["J.C.P. Miller"],
    title="Numerical Quadrature Over a Rectangular Domain in Two or More Dimensions. Part 3: Quadrature of a Harmonic Integrand",
    journal="Mathematics of Computation",
    volume="14",
    number="71",
    month="jul",
    year="1960",
    pages="240-248",
    url="https://doi.org/10.2307/2003163",
)


def miller():
    d = {
        "zero2": [[frac(250, 225)]],
        "d4_a0": [[-frac(8, 225)], [1]],
        "d4_aa": [[frac(7, 900)], [1]],
    }
    # This scheme is exact for _harmonic_ integrands of degree <= 11.
    return C2Scheme("Miller", d, 1, source)


register([miller])



import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["Zhongxuan Luo", "Zhaoliang Meng"],
    title="Cubature formulas over the n-sphere",
    journal="Journal of Computational and Applied Mathematics",
    year="2007",
    volume="202",
    pages="511-522",
    url="https://doi.org/10.1016/j.cam.2006.03.004",
)

this_dir = pathlib.Path(__file__).resolve().parent


def luo_meng_1():
    return _read(this_dir / "luo_meng_1.json", _source)


def luo_meng_2():
    return _read(this_dir / "luo_meng_2.json", _source)


def luo_meng_3():
    return _read(this_dir / "luo_meng_3.json", _source)


def luo_meng_4():
    return _read(this_dir / "luo_meng_4.json", _source)


def luo_meng_5():
    return _read(this_dir / "luo_meng_5.json", _source)


def luo_meng_6():
    return _read(this_dir / "luo_meng_6.json", _source)


# TODO find error
# def luo_meng_7():
#     return _read(this_dir / "luo_meng_7.json", _source)


register([luo_meng_1, luo_meng_2, luo_meng_3, luo_meng_4, luo_meng_5, luo_meng_6])



from __future__ import annotations

import json
import warnings

import numpy as np

from .._exception import QuadpyError
from ..helpers import QuadratureScheme, expand_symmetries, plot_disks

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class S2Scheme(QuadratureScheme):
    def __init__(
        self,
        name: str,
        symmetry_data,
        degree: int,
        source=None,
        tol: float = 1.0e-14,
        weight_factor: float | None = None,
    ):
        self.domain = "S2"
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data, dim=2)
        assert points.shape[0] == 2
        assert points.shape[1] == weights.shape[0], f"{points.shape}, {weights.shape}"
        if weight_factor is not None:
            weights *= weight_factor
        super().__init__(name, weights, points, degree, source, tol)

    def plot(self, show_axes=False):
        from matplotlib import pyplot as plt

        ax = plt.gca()
        # change default range so that new disks will work
        plt.axis("equal")
        ax.set_xlim((-1.5, 1.5))
        ax.set_ylim((-1.5, 1.5))

        if not show_axes:
            ax.set_axis_off()

        disk1 = plt.Circle((0, 0), 1, color="k", fill=False)
        ax.add_patch(disk1)

        plot_disks(plt, self.points.T, self.weights, np.pi)

    def integrate(self, f, center, radius, dot=np.dot):
        center = np.array(center)
        rr = np.multiply.outer(radius, self.points.T)
        rr = np.swapaxes(rr, 0, -2)
        x = (rr + center).T
        fx = np.array(f(x))
        if fx.shape[-len(x.shape[1:]) :] != x.shape[1:]:
            string = ", ".join(str(val) for val in x.shape[1:])
            raise QuadpyError(
                f"Wrong return value shape {fx.shape}. Expected (..., {string})."
            )
        return np.pi * np.array(radius) ** 2 * dot(fx, self.weights)


def _read(filepath, source) -> S2Scheme:
    with open(filepath) as f:
        content = json.load(f)

    degree = content["degree"]
    name = content["name"]
    tol = content["test_tolerance"]

    if "_ERR" in content:
        warnings.warn(content["_ERR"])

    weight_factor = content["weight factor"] if "weight factor" in content else None

    return S2Scheme(
        name, content["data"], degree, source, tol, weight_factor=weight_factor
    )


def _scheme_from_dict(content, source=None):
    points, weights = expand_symmetries(content["data"])

    if "weight factor" in content:
        weights *= content["weight factor"]

    return S2Scheme(
        content["name"],
        weights,
        points,
        degree=content["degree"],
        source=source,
        tol=content["test_tolerance"],
        comments=content["comments"] if "comments" in content else None,
    )


def get_good_scheme(degree: int) -> S2Scheme | None:
    if degree <= 19:
        return {
            0: schemes["midpoint"],
            1: schemes["midpoint"],
            2: schemes["albrecht_collatz"],
            3: schemes["albrecht_collatz"],
            4: schemes["mysovskih_1"],
            5: schemes["radon"],
            6: schemes["kim_song_6"],
            7: schemes["kim_song_6"],
            8: schemes["luo_meng_2"],
            9: schemes["luo_meng_2"],
            10: schemes["mysovskih_2"],
            11: schemes["mysovskih_2"],
            12: schemes["cools_haegemans_13_1"],
            13: schemes["cools_haegemans_13_1"],
            14: schemes["kim_song_11"],
            15: schemes["kim_song_11"],
            16: schemes["cools_kim_1"],
            17: schemes["cools_kim_1"],
            18: schemes["kim_song_15"],
            19: schemes["kim_song_15"],
        }[degree]()

    return None



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["Johannes W. Wissmann", "Thomas Becker"],
    title="Partially Symmetric Cubature Formulas for Even Degrees of Exactness",
    journal="SIAM J. Numer. Anal.",
    volume="23",
    number="3",
    pages="676–685",
    url="https://doi.org/10.1137/0723043",
)

this_dir = pathlib.Path(__file__).resolve().parent


def wissmann_becker_4_1():
    return _read(this_dir / "wissmann_becker_4_1.json", source)


def wissmann_becker_4_2():
    return _read(this_dir / "wissmann_becker_4_2.json", source)


def wissmann_becker_6_1():
    return _read(this_dir / "wissmann_becker_6_1.json", source)


def wissmann_becker_6_2():
    return _read(this_dir / "wissmann_becker_6_2.json", source)


def wissmann_becker_8_1():
    return _read(this_dir / "wissmann_becker_8_1.json", source)


def wissmann_becker_8_2():
    return _read(this_dir / "wissmann_becker_8_2.json", source)


register(
    [
        wissmann_becker_4_1,
        wissmann_becker_4_2,
        wissmann_becker_6_1,
        wissmann_becker_6_2,
        wissmann_becker_8_1,
        wissmann_becker_8_2,
    ]
)



import numpy as np

from ..helpers import QuadratureScheme, plot_disks

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class U2Scheme(QuadratureScheme):
    def __init__(self, name, source, degree, weights, points, tol):
        super().__init__(name, weights, points, degree, source, tol)
        self.domain = "U2"

    def show(self, *args, **kwargs):
        from matplotlib import pyplot as plt

        self.plot(*args, **kwargs)
        plt.show()

    def plot(self, show_axes=False):
        from matplotlib import pyplot as plt

        ax = plt.gca()
        # change default range so that new disks will work
        ax.axis("equal")
        ax.set_xlim((-1.5, 1.5))
        ax.set_ylim((-1.5, 1.5))

        if not show_axes:
            ax.set_axis_off()

        disk1 = plt.Circle((0, 0), 1, color="k", fill=False)
        ax.add_patch(disk1)

        # The total area is used to gauge the disk radii. This is only meaningful for 2D
        # manifolds, not for the circle. What we do instead is choose the total_area
        # such that the sum of the disk radii equals pi.
        total_area = np.pi ** 3 / len(self.weights)
        plot_disks(plt, self.points, self.weights, total_area)

    def integrate(self, f, center, radius, dot=np.dot):
        center = np.array(center)
        rr = np.multiply.outer(radius, self.points)
        rr = np.swapaxes(rr, 0, -2)
        ff = np.array(f((rr + center).T))
        ref_vol = 2 * np.pi * np.asarray(radius)
        return ref_vol * dot(ff, self.weights)


def get_good_scheme(n):
    return schemes["krylov"](n)



import pathlib

from sympy import Rational as frac

from ...helpers import article
from .._helpers import T2Scheme, _read, register

source = article(
    authors=["D.A. Dunavant"],
    title="High Degree Efficient Symmetrical Gaussian Quadrature Rules for the Triangle",
    journal="Article in International Journal for Numerical Methods in Engineering",
    volume="21",
    number="6",
    pages="1129-1148",
    month="jun",
    year="1985",
    url="https://doi.org/10.1002/nme.1620210612",
)

this_dir = pathlib.Path(__file__).resolve().parent


def dunavant_01():
    d = {"centroid": [[1]]}
    return T2Scheme("Dunavant 1", d, 1, source, 7.850e-17)


def dunavant_02():
    d = {"d3_aa": [[frac(1, 3)], [frac(1, 6)]]}
    return T2Scheme("Dunavant 2", d, 2, source, 2.220e-16)


def dunavant_03():
    d = {"centroid": [[-frac(9, 16)]], "d3_aa": [[frac(25, 48)], [frac(1, 5)]]}
    return T2Scheme("Dunavant 3", d, 3, source, 6.661e-16)


def dunavant_04():
    return _read(this_dir / "dunavant_04.json", source)


def dunavant_05():
    return _read(this_dir / "dunavant_05.json", source)


def dunavant_06():
    return _read(this_dir / "dunavant_06.json", source)


def dunavant_07():
    return _read(this_dir / "dunavant_07.json", source)


def dunavant_08():
    return _read(this_dir / "dunavant_08.json", source)


def dunavant_09():
    # DUP equals TRIEX 19
    return _read(this_dir / "dunavant_09.json", source)


def dunavant_10():
    return _read(this_dir / "dunavant_10.json", source)


def dunavant_11():
    return _read(this_dir / "dunavant_11.json", source)


def dunavant_12():
    return _read(this_dir / "dunavant_12.json", source)


def dunavant_13():
    return _read(this_dir / "dunavant_13.json", source)


def dunavant_14():
    return _read(this_dir / "dunavant_14.json", source)


def dunavant_15():
    return _read(this_dir / "dunavant_15.json", source)


def dunavant_16():
    return _read(this_dir / "dunavant_16.json", source)


def dunavant_17():
    return _read(this_dir / "dunavant_17.json", source)


def dunavant_18():
    return _read(this_dir / "dunavant_18.json", source)


def dunavant_19():
    return _read(this_dir / "dunavant_19.json", source)


def dunavant_20():
    return _read(this_dir / "dunavant_20.json", source)


register(
    [
        dunavant_01,
        dunavant_02,
        dunavant_03,
        dunavant_04,
        dunavant_05,
        dunavant_06,
        dunavant_07,
        dunavant_08,
        dunavant_09,
        dunavant_10,
        dunavant_11,
        dunavant_12,
        dunavant_13,
        dunavant_14,
        dunavant_15,
        dunavant_16,
        dunavant_17,
        dunavant_18,
        dunavant_19,
        dunavant_20,
    ]
)



import pathlib

from ...helpers import article
from .._helpers import _read, register

source = article(
    authors=["R. Cools", "Ann Haegemans"],
    title="Another Step Forward in Searching for Cubature Formulae with a Minimal Number of Knots for the Square",
    journal="Computing",
    volume="40",
    pages="139-146",
    year="1988",
    url="https://doi.org/10.1007/BF02247942",
)

this_dir = pathlib.Path(__file__).resolve().parent


def cools_haegemans_1988_1():
    return _read(this_dir / "cools_haegemans_1988_1.json", source)


def cools_haegemans_1988_2():
    return _read(this_dir / "cools_haegemans_1988_2.json", source)


register([cools_haegemans_1988_1, cools_haegemans_1988_2])



import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["R. Cools", "K.J. Kim"],
    title="A survey of known and new cubature formulas for the unit disk",
    journal="Korean J. Comput. & Appl. Math.",
    volume="7",
    month="sep",
    year="2000",
    number="3",
    pages="477-485",
    # url="https://doi.org/10.1007/BF03012263"
)

this_dir = pathlib.Path(__file__).resolve().parent


def cools_kim_1():
    return _read(this_dir / "cools_kim_1.json", _source)


def cools_kim_2():
    return _read(this_dir / "cools_kim_2.json", _source)


def cools_kim_3():
    return _read(this_dir / "cools_kim_3.json", _source)


register([cools_kim_1, cools_kim_2, cools_kim_3])



import numpy as np
from sympy import sqrt

from ..helpers import article, untangle
from ._helpers import CnScheme, _s

_source = article(
    authors=["Henry C. Thacher"],
    title="An efficient composite formula for multidimensional quadrature",
    journal="Communications of the ACM",
    volume="7",
    number="1",
    month="jan",
    year="1964",
    pages="23-25",
    url="https://doi.org/10.1145/363872.363897",
)


def thacher(n):
    r = sqrt(3) / 6
    data = [(1, [n * [2 * r]]), (+r, _s(n, -1, r)), (-r, _s(n, +1, r))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return CnScheme(f"Thacher ({n}D)", n, weights, points, 2, _source, 1.399e-14)



import warnings

import ndim

from ..cn._mcnamee_stenger import (
    _mcnamee_stenger_3,
    _mcnamee_stenger_5,
    _mcnamee_stenger_7,
    _mcnamee_stenger_9,
)
from ._helpers import EnrScheme


def integrator(n, k, symbolic):
    """Returns the integral of the polynomial given by the coefficients k over Enr."""
    alpha = k + [0] * (n - len(k))
    return ndim.enr.integrate_monomial(alpha, symbolic)


def mcnamee_stenger_3(n, symbolic=False):
    return EnrScheme(*_mcnamee_stenger_3(n, integrator, symbolic=symbolic))


def mcnamee_stenger_5(n, symbolic=False):
    return EnrScheme(*_mcnamee_stenger_5(n, integrator, symbolic=symbolic))


def mcnamee_stenger_7a(n, symbolic=False):
    return EnrScheme(
        *_mcnamee_stenger_7(n, integrator, False, symbolic=symbolic), 7.530e-14
    )


def mcnamee_stenger_7b(n, symbolic=False):
    return EnrScheme(*_mcnamee_stenger_7(n, integrator, True, symbolic=symbolic))


def mcnamee_stenger_9a(n, symbolic=False):
    scheme = EnrScheme(*_mcnamee_stenger_9(n, integrator, False, symbolic=symbolic))
    warnings.warn(f"{scheme.name} is very badly conditioned.")
    return scheme


def mcnamee_stenger_9b(n, symbolic=False):
    scheme = EnrScheme(*_mcnamee_stenger_9(n, integrator, True, symbolic=symbolic))
    warnings.warn(f"{scheme.name} is very badly conditioned.")
    return scheme



import math

import sympy

from ..helpers import article
from ._helpers import C3Scheme, register

_source = article(
    authors=["Preston C. Hammer", "A. Wayne Wymore"],
    title="Numerical evaluation of multiple integrals. I",
    journal="Math. Comp.",
    volume="11",
    year="1957",
    pages="59-67",
    url="https://doi.org/10.1090/S0025-5718-1957-0087220-6",
)


# hw(27 / 20) has all the desirable properties (positive weights, points inside the
# domain) while minimizing the ratio max(weights) / min(weights)
def hammer_wymore(lmbda=sympy.Rational(27, 20)):
    symbolic = not isinstance(lmbda, float)
    frac = sympy.Rational if symbolic else lambda x, y: x / y
    sqrt = sympy.sqrt if symbolic else math.sqrt

    I0 = 8
    I2 = frac(8, 3)
    I4 = frac(8, 5)
    I22 = frac(8, 9)
    I6 = frac(8, 7)
    I42 = frac(8, 15)
    I222 = frac(8, 27)

    u1 = (I6 - 2 * I42 + I222 + lmbda * (I42 - I222)) / (I4 - I22)
    u2 = u1 / lmbda
    a1 = (I6 - 2 * I42 + I222) / 2 / u1 ** 3
    a2 = (I42 - I222) / 4 / u2 ** 3

    p0 = I0 - 6 * a1 - 12 * a2
    p1 = I2 - 2 * a1 * u1 - 8 * a2 * u2
    p2 = I22 - 4 * a2 * u2 ** 2
    p3 = I222

    c0 = (p3 * p1 - p2 ** 2) / (p0 * p2 - p1 ** 2)
    c1 = (p3 * p0 - p1 * p2) / (p1 ** 2 - p2 * p0)

    b = sqrt(c1 ** 2 - 4 * c0)
    u3 = (-c1 - b) / 2
    u4 = (-c1 + b) / 2

    a3 = (p0 * u4 - p1) / 8 / (u4 - u3)
    a4 = (p0 * u3 - p1) / 8 / (u3 - u4)

    x1 = sqrt(u1)
    x2 = sqrt(u2)
    x3 = sqrt(u3)
    x4 = sqrt(u4)

    d = {
        "symm_r00": [[a1 / 8], [x1]],
        "symm_rr0": [[a2 / 8], [x2]],
        "symm_rrr": [[a3 / 8, a4 / 8], [x3, x4]],
    }
    return C3Scheme(f"Hammer-Wymore (lambda = {lmbda})", d, 7, _source)


register([hammer_wymore])



import ndim
import numpy as np

from .._exception import QuadpyError
from ..helpers import QuadratureScheme


class SnScheme(QuadratureScheme):
    def __init__(self, name, dim, weights, points, degree, source, tol=1.0e-14):
        self.domain = f"Sn (n={dim})"
        self.dim = dim
        assert points.shape[0] == dim
        super().__init__(name, weights, points, degree, source, tol)

    def points_inside(self):
        return np.einsum("ij,ij->i", self.points, self.points) < 1

    def points_inside_or_boundary(self):
        return np.einsum("ij,ij->i", self.points, self.points) <= 1

    def integrate(self, f, center, radius, dot=np.dot):
        center = np.array(center)
        rr = np.multiply.outer(radius, self.points.T)
        rr = np.swapaxes(rr, 0, -2)
        x = (rr + center).T
        fx = np.array(f(x))
        if fx.shape[-len(x.shape[1:]) :] != x.shape[1:]:
            string = ", ".join(str(val) for val in x.shape[1:])
            raise QuadpyError(
                f"Wrong return value shape {fx.shape}. Expected (..., {string})."
            )
        ref_vol = ndim.nball.volume(self.dim, r=np.asarray(radius), symbolic=False)
        return ref_vol * dot(fx, self.weights)



import numpy as np
from sympy import Rational as frac

from ..helpers import book, untangle
from ._hammer_stroud import hammer_stroud_1a as stroud_tn_2_1a
from ._hammer_stroud import hammer_stroud_1b as stroud_tn_2_1b
from ._hammer_stroud import hammer_stroud_2 as stroud_tn_3_1
from ._helpers import TnScheme
from ._lauffer import lauffer_1 as stroud_tn_1_2
from ._lauffer import lauffer_2 as stroud_tn_2_2
from ._lauffer import lauffer_3 as stroud_tn_3_9
from ._lauffer import lauffer_4 as stroud_tn_4_1
from ._lauffer import lauffer_5 as stroud_tn_5_2
from ._stroud_1961 import stroud_1961 as stroud_tn_3_3
from ._stroud_1964 import stroud_1964a as stroud_tn_3_6a
from ._stroud_1964 import stroud_1964b as stroud_tn_3_6b
from ._stroud_1966 import stroud_1966_1 as stroud_tn_3_2
from ._stroud_1966 import stroud_1966_2 as stroud_tn_3_4
from ._stroud_1966 import stroud_1966_3 as stroud_tn_3_5
from ._stroud_1966 import stroud_1966_4 as stroud_tn_3_7
from ._stroud_1966 import stroud_1966_5 as stroud_tn_3_8
from ._stroud_1966 import stroud_1966_6 as stroud_tn_3_10
from ._stroud_1966 import stroud_1966_7 as stroud_tn_3_11
from ._stroud_1969 import stroud_1969 as stroud_tn_5_1

source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_tn_1_1(n):
    # midpoint rule
    degree = 1
    data = [(1, np.full((1, n + 1), frac(1, n + 1)))]
    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud Tn 1-1", n, weights, points, degree, source)


__all__ = [
    "stroud_tn_1_1",
    "stroud_tn_1_2",
    "stroud_tn_2_1a",
    "stroud_tn_2_1b",
    "stroud_tn_2_2",
    "stroud_tn_3_1",
    "stroud_tn_3_2",
    "stroud_tn_3_3",
    "stroud_tn_3_4",
    "stroud_tn_3_5",
    "stroud_tn_3_6a",
    "stroud_tn_3_6b",
    "stroud_tn_3_7",
    "stroud_tn_3_8",
    "stroud_tn_3_9",
    "stroud_tn_3_10",
    "stroud_tn_3_11",
    "stroud_tn_4_1",
    "stroud_tn_5_1",
    "stroud_tn_5_2",
]



import numpy as np

from ..helpers import QuadratureScheme, plot_disks_1d


class E1r2Scheme(QuadratureScheme):
    def __init__(self, name, weights, points, degree, source=None):
        self.domain = "E1r2"
        super().__init__(name, weights, points, degree, source)

    def integrate(self, f, dot=np.dot):
        x = np.array([self.points.T])
        fx = np.asarray(f(x))
        return dot(fx, self.weights)

    def plot(self, show_axes=False):
        from matplotlib import pyplot as plt

        if not show_axes:
            plt.gca().set_axis_off()

        plt.axis("equal")
        m = 1.1 * np.max(self.points)
        plt.plot([-m, +m], [0, 0], color="k")
        pts = np.column_stack([self.points, np.zeros(len(self.points))])
        plot_disks_1d(plt, pts, self.weights, total_area=1.0)



from __future__ import annotations

from typing import Callable

import numpy as np

from ._gauss_kronrod import _gauss_kronrod_integrate


def _numpy_all_except_last(a):
    return np.all(a, axis=tuple(range(len(a.shape) - 1)))


class IntegrationError(Exception):
    pass


def integrate_adaptive(
    f: Callable,
    intervals,
    eps_abs: float | None = 1.0e-10,
    eps_rel: float | None = 1.0e-10,
    criteria_connection: Callable = np.any,
    # Use 21-point Gauss-Kronrod like QUADPACK
    # <https://en.wikipedia.org/wiki/QUADPACK#General-purpose_routines>
    kronrod_degree: int = 10,
    minimum_interval_length: float = 0.0,
    max_num_subintervals=np.inf,
    dot: Callable = np.dot,
    domain_shape: tuple | None = None,
    range_shape: tuple | None = None,
):
    intervals = np.asarray(intervals)
    assert intervals.shape[0] == 2
    assert np.all(intervals > -np.inf)
    assert np.all(intervals < np.inf)

    assert (
        eps_abs is not None or eps_rel is not None
    ), "One of eps_abs, eps_rel must be specified."

    # Use Gauss-Kronrod scheme for error estimation and adaptivity.
    # The method also returns guesses for the domain_shape and range_shape (if any of
    # them is None).
    out = _gauss_kronrod_integrate(
        kronrod_degree,
        f,
        intervals,
        dot=dot,
        domain_shape=domain_shape,
        range_shape=range_shape,
    )
    value_estimates = out.val_gauss_legendre
    interval_lengths = out.interval_lengths
    error_estimates = out.error_estimate
    domain_shape = out.domain_shape
    range_shape = out.range_shape

    # Flatten the list of intervals so we can do good-bad bookkeeping via a list.
    intervals = intervals.reshape((2,) + domain_shape + (-1,))
    num_subintervals = 1
    orig_shape = value_estimates.shape
    value_estimates = value_estimates.reshape(range_shape + (-1,))
    error_estimates = error_estimates.reshape(range_shape + (-1,))

    # Mark intervals with acceptable approximations. For this, take all() across every
    # dimension except the last one (which is the interval index).
    criteria = []
    if eps_abs is not None:
        is_okay = error_estimates < eps_abs
        criteria.append(_numpy_all_except_last(is_okay))
    if eps_rel is not None:
        is_okay = error_estimates < eps_rel * np.abs(value_estimates)
        criteria.append(_numpy_all_except_last(is_okay))
    is_good = criteria_connection(criteria, axis=0)

    good_values_sum = np.sum(value_estimates[..., is_good], axis=-1)
    good_errors_sum = np.sum(error_estimates[..., is_good], axis=-1)

    while np.any(~is_good):
        # split the bad intervals in half
        bad_intervals = intervals[..., ~is_good]
        midpoints = 0.5 * (bad_intervals[0] + bad_intervals[1])
        intervals = np.array(
            [
                np.concatenate([bad_intervals[0], midpoints], axis=-1),
                np.concatenate([midpoints, bad_intervals[1]], axis=-1),
            ]
        )
        # idx = np.concatenate([idx[~is_good], idx[~is_good]])
        num_subintervals += np.sum(is_good)

        if num_subintervals > max_num_subintervals:
            raise IntegrationError(
                f"Tolerances (abs: {eps_abs}, rel: {eps_rel}) could not be reached "
                f"with the given max_num_subintervals (= {max_num_subintervals})."
            )

        # compute values and error estimates for the new intervals
        out = _gauss_kronrod_integrate(
            kronrod_degree,
            f,
            intervals,
            dot=dot,
            domain_shape=domain_shape,
            range_shape=range_shape,
        )
        value_estimates = out.val_gauss_legendre
        interval_lengths = out.interval_lengths
        error_estimates = out.error_estimate

        # mark good intervals, gather values and error estimates
        if np.any(interval_lengths < minimum_interval_length):
            raise IntegrationError(
                f"Tolerances (abs: {eps_abs}, rel: {eps_rel}) could not be reached "
                f"with the given minimum_interval_length (= {minimum_interval_length})."
            )

        # tentative total value (as if all intervals were good)
        ttv = good_values_sum + np.sum(value_estimates, axis=-1)

        # distribute the remaining allowances according to the interval lengths
        tau = interval_lengths / np.sum(interval_lengths)

        criteria = []
        if eps_abs is not None:
            allowance_abs = eps_abs - good_errors_sum
            is_okay = error_estimates < np.multiply.outer(allowance_abs, tau)
            criteria.append(_numpy_all_except_last(is_okay))
        if eps_rel is not None:
            # allowance_rel = eps_rel - good_errors_sum / np.abs(ttv)
            # is_okay = error_estimates < tau * allowance_rel * np.abs(ttv)
            allowance_rel_ttv = eps_rel * np.abs(ttv) - good_errors_sum
            is_okay = error_estimates < np.multiply.outer(allowance_rel_ttv, tau)
            criteria.append(_numpy_all_except_last(is_okay))
        is_good = criteria_connection(criteria, axis=0)

        good_values_sum += np.sum(value_estimates[..., is_good], axis=-1)
        good_errors_sum += np.sum(error_estimates[..., is_good], axis=-1)

    good_values_sum = good_values_sum.reshape(orig_shape)
    good_errors_sum = good_errors_sum.reshape(orig_shape)
    return good_values_sum, good_errors_sum



from ..helpers import article
from ._hammer_wymore import hammer_wymore
from ._helpers import C3Scheme, register

source = article(
    authors=["V.L.N. Sarma", "A. H. Stroud"],
    title="Eberlein Measure and Mechanical Quadrature Formulae. II. Numerical Results",
    journal="Mathematics of Computation",
    volume="23",
    number="108",
    month="oct",
    year="1969",
    pages="781-784",
    url="https://doi.org/10.2307/2004963",
)


def sarma_stroud():
    # Hammer-Wymore is a one-parameter family of schemes, and the parameter lambda is
    # chosen to minimize the standard deviation of Sarma's error functional. The
    # particular value of lambda is not explicitly given in the article, but computed
    # from the specified values. Note that it is only given in single precision.
    hw = hammer_wymore(lmbda=1.0329785305)
    return C3Scheme("Sarma-Stroud", hw.symmetry_data, hw.degree, source)


register([sarma_stroud])



from __future__ import annotations

import json

import numpy as np

from ..cn import CnScheme
from ..cn import ncube_points as rectangle_points
from ..cn import transform
from ..helpers import expand_symmetries, plot_disks
from ..tn import get_vol

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class C2Scheme(CnScheme):
    def __init__(
        self,
        name: str,
        symmetry_data,
        degree: int,
        source=None,
        tol: float = 1.0e-14,
        comments: list[str] | None = None,
    ):
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data, dim=2)
        super().__init__(name, 2, weights, points, degree, source, tol, comments)
        self.domain = "C2"

    def plot(self, quad=rectangle_points([0.0, 1.0], [0.0, 1.0]), show_axes=False):
        """Shows the quadrature points on a given quad. The area of the disks around the
        points coincides with their weights.
        """
        from matplotlib import pyplot as plt

        def plot_segment(a, b):
            plt.plot((a[0], b[0]), (a[1], b[1]), "-k")

        plot_segment(quad[0][0], quad[1][0])
        plot_segment(quad[1][0], quad[1][1])
        plot_segment(quad[1][1], quad[0][1])
        plot_segment(quad[0][1], quad[0][0])

        if not show_axes:
            plt.gca().set_axis_off()

        transformed_pts = transform(self.points, quad)

        # compute volume by splitting it in two triangles
        vol = get_vol(np.array([quad[0][0], quad[1][0], quad[0][1]])) + get_vol(
            np.array([quad[0][0], quad[0][1], quad[1][1]])
        )
        plot_disks(plt, transformed_pts, self.weights, vol)
        plt.axis("equal")
        plt.xlim(-0.1, 1.1)
        plt.ylim(-0.1, 1.1)


def _read(filepath, source):
    with open(filepath) as f:
        content = json.load(f)

    degree = content["degree"]
    name = content["name"]
    tol = content["test_tolerance"]

    data = content["data"]
    if "weight factor" in content:
        w = content["weight factor"]
        for value in data.values():
            value[0] = [val * w for val in value[0]]

    return C2Scheme(name, data, degree, source, tol)


def _scheme_from_dict(content, source=None):
    points, weights = expand_symmetries(content["data"])

    if "weight factor" in content:
        weights *= content["weight factor"]

    return C2Scheme(
        content["name"],
        weights,
        points,
        degree=content["degree"],
        source=source,
        tol=content["test_tolerance"],
        comments=content["comments"] if "comments" in content else None,
    )


def get_good_scheme(degree: int) -> C2Scheme | None:
    if degree <= 22:
        return {
            0: schemes["dunavant_00"],
            1: schemes["dunavant_00"],
            2: schemes["hammer_stroud_1_2"],
            3: schemes["hammer_stroud_1_2"],
            4: schemes["burnside"],
            5: schemes["burnside"],
            6: schemes["tyler_2"],
            7: schemes["tyler_2"],
            8: schemes["rabinowitz_richter_1"],
            9: schemes["rabinowitz_richter_1"],
            10: schemes["witherden_vincent_11"],
            11: schemes["witherden_vincent_11"],
            12: schemes["witherden_vincent_13"],
            13: schemes["witherden_vincent_13"],
            14: schemes["rabinowitz_richter_6"],
            15: schemes["rabinowitz_richter_6"],
            16: schemes["witherden_vincent_17"],
            17: schemes["witherden_vincent_17"],
            18: schemes["witherden_vincent_19"],
            19: schemes["witherden_vincent_19"],
            20: schemes["witherden_vincent_21"],
            21: schemes["witherden_vincent_21"],
        }[degree]()

    return None



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, fsd, pm, untangle
from ._helpers import UnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Some Seventh Degree Integration Formulas for Symmetric Regions",
    journal="SIAM J. Numer. Anal.",
    volume="4",
    number="1",
    pages="37–44",
    url="https://doi.org/10.1137/0704004",
)


def stroud_1967(n):
    degree = 7

    r = 1
    s = sqrt(frac(1, n))
    t = sqrt(frac(1, 2))
    B = frac(8 - n, n * (n + 2) * (n + 4))
    C = frac(n ** 3, 2 ** n * n * (n + 2) * (n + 4))
    D = frac(4, n * (n + 2) * (n + 4))

    data = [
        (B, fsd(n, (r, 1))),
        (C, pm(n * [s])),
        # ERR Stroud's book incorrectly states (t, t,..., t)_FS instead of
        # (t, t, 0, ..., 0)_FS.
        (D, fsd(n, (t, 2))),
    ]

    points, weights = untangle(data)
    return UnScheme("Stroud 1967", n, weights, points, degree, source)



import numpy as np
import sympy

from ..helpers import phdthesis
from ._helpers import S2Scheme, register

_source = phdthesis(
    authors=["William Hollis Peirce"],
    title="Numerical integration over planar regions",
    school="University of Wisconsin--Madison",
    year="1956",
    url="https://books.google.de/books/about/Numerical_integration_over_planar_region.html?id=WR9SAAAAMAAJ",
)

frac = sympy.Rational
pi = sympy.pi
sqrt = np.vectorize(sympy.sqrt)
cos = np.vectorize(sympy.cos)
sin = np.vectorize(sympy.sin)
pm_ = np.array([+1, -1])


def peirce_1956_1():
    sqrt29 = sqrt(29)
    r = sqrt(frac(3, 4))
    s, t = sqrt((27 - pm_ * 3 * sqrt29) / 104)

    B1 = frac(2, 27)
    # ERR Stroud incorrectly lists 4 instead of 41 here.
    B2, B3 = (551 + pm_ * 41 * sqrt29) / 6264

    d = {"d4_a0": [[B1], [r]], "d4_aa": [[B2, B3], [s, t]]}
    return S2Scheme("Peirce 1956-1", d, 7, _source)


def peirce_1956_2():
    sqrt15 = sqrt(15)
    cos_pi8 = cos(pi / 8)
    sin_pi8 = sin(pi / 8)

    r = sqrt((5 + sqrt15) / 10)
    u1 = sqrt((5 - sqrt15) / 10) * cos_pi8
    v1 = sqrt((5 - sqrt15) / 10) * sin_pi8
    u2 = sqrt(frac(1, 2)) * cos_pi8
    v2 = sqrt(frac(1, 2)) * sin_pi8
    u3, v3 = sqrt((5 + sqrt15 + pm_ * sqrt(185 * sqrt15 - 700)) / 20)

    B1 = (12060 - 1440 * sqrt15) / 254088
    B2 = frac(5, 144)
    B3 = frac(1, 18)
    B4 = (5585 + 1440 * sqrt15) / 508176

    d = {
        "d4_a0": [[B1], [r]],
        "d4_ab": [[B2, B3, B4], [u1, u2, u3], [v1, v2, v3]],
    }
    return S2Scheme("Peirce 1956-2", d, 9, _source)


def peirce_1956_3():
    sqrt15 = sqrt(15)

    B1 = frac(5, 144)
    B2 = (34 - 5 * sqrt15) / 396
    B3 = (4805 - 620 * sqrt15) / 103824
    C1 = (10 + 5 * sqrt15) / 792
    C2 = (2405 + 620 * sqrt15) / 207648
    D = B1

    r1 = sqrt((5 - sqrt15) / 10)
    r2 = sqrt(frac(1, 2))
    r3 = sqrt((5 + sqrt15) / 10)
    u1, v1 = sqrt((5 + pm_ * sqrt(45 - 10 * sqrt15)) / 20)
    u2, v2 = sqrt((5 + sqrt15 + pm_ * 2 * sqrt(40 * sqrt15 - 150)) / 20)
    t = sqrt((5 - sqrt15) / 20)

    d = {
        "d4_a0": [[B1, B2, B3], [r1, r2, r3]],
        "d4_ab": [[C1, C2], [u1, u2], [v1, v2]],
        "d4_aa": [[D], [t]],
    }
    return S2Scheme("Peirce 1956-3", d, 11, _source)


register([peirce_1956_1, peirce_1956_2, peirce_1956_3])



import numpy as np

from ..helpers import article
from ._helpers import C1Scheme

source = article(
    authors=["J. Waldvogel"],
    title="Fast Construction of the Fejér and Clenshaw–Curtis Quadrature Rules",
    journal="BIT Numerical Mathematics",
    month="mar",
    year="2006",
    volume="46",
    number="1",
    pages="195–202",
    url="https://doi.org/10.1007/s10543-006-0045-4",
)


def clenshaw_curtis(n):
    degree = n

    points = -np.cos((np.pi * np.arange(n)) / (n - 1))

    if n == 2:
        weights = np.array([1.0, 1.0])
        return C1Scheme("Clenshaw-Curtis", degree, weights, points)

    n -= 1
    N = np.arange(1, n, 2)
    length = len(N)
    m = n - length
    v0 = np.concatenate([2.0 / N / (N - 2), np.array([1.0 / N[-1]]), np.zeros(m)])
    v2 = -v0[:-1] - v0[:0:-1]
    g0 = -np.ones(n)
    g0[length] += n
    g0[m] += n
    g = g0 / (n ** 2 - 1 + (n % 2))

    w = np.fft.ihfft(v2 + g)
    assert max(w.imag) < 1.0e-15
    w = w.real

    if n % 2 == 1:
        weights = np.concatenate([w, w[::-1]])
    else:
        weights = np.concatenate([w, w[len(w) - 2 :: -1]])

    return C1Scheme("Clenshaw-Curtis", degree, weights, points, source)



import pathlib

from sympy import Rational as frac
from sympy import sqrt

from ...helpers import article
from .._helpers import T3Scheme, _read, register

source = article(
    authors=["Lee Shunn", "Frank Ham"],
    title="Symmetric quadrature rules for tetrahedra based on a cubic close-packed lattice arrangement",
    journal="Journal of Computational and Applied Mathematics",
    year="2012",
    url="https://doi.org/10.1016/j.cam.2012.03.032",
)

this_dir = pathlib.Path(__file__).resolve().parent


def shunn_ham_1():
    d = {"s4": [[1]]}
    return T3Scheme("Shunn-Ham 2", d, 1, source)


def shunn_ham_2():
    d = {"s31": [[frac(1, 4)], [frac(1, 4) - sqrt(5) / 20]]}
    return T3Scheme("Shunn-Ham 2", d, 2, source)


def shunn_ham_3():
    return _read(this_dir / "shunn_ham_3.json", source)


def shunn_ham_4():
    return _read(this_dir / "shunn_ham_4.json", source)


def shunn_ham_5():
    return _read(this_dir / "shunn_ham_5.json", source)


def shunn_ham_6():
    return _read(this_dir / "shunn_ham_6.json", source)


register([shunn_ham_1, shunn_ham_2, shunn_ham_3, shunn_ham_4, shunn_ham_5, shunn_ham_6])



import ndim
import numpy as np
from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article, rd, untangle
from ._helpers import TnScheme

source = article(
    authors=["A.H. Stroud"],
    title="A Fifth Degree Integration Formula for the n-Simplex",
    journal="SIAM J. Numer. Anal.",
    volume="6",
    number="1",
    pages="90–98",
    url="https://doi.org/10.1137/0706009",
)


def stroud_1969(n):
    assert n >= 3

    degree = 5

    sqrt15 = sqrt(15)

    t = frac(1, n + 1)
    r1, r2 = ((n + 4 - pm * sqrt15) / (n ** 2 + 8 * n + 1) for pm in [+1, -1])
    s1, s2 = ((4 * n + 1 + pm * n * sqrt15) / (n ** 2 + 8 * n + 1) for pm in [+1, -1])
    u1, u2 = ((n + 7 + pm * 2 * sqrt15) / (n ** 2 + 14 * n - 11) for pm in [+1, -1])
    v1, v2 = (
        (4 * n - 2 - pm * (n - 1) * sqrt15) / (n ** 2 + 14 * n - 11) for pm in [+1, -1]
    )

    # Solve linear equation system for x^k, k={0, 2, 3, 4, 5}, for the
    # weights (the same is done in Stroud's article).
    pts = [
        np.full((1, n + 1), t),
        rd(n + 1, [(r1, n), (s1, 1)]),
        rd(n + 1, [(r2, n), (s2, 1)]),
        rd(n + 1, [(u1, n - 1), (v1, 2)]),
    ]
    k_range = [0, 2, 3, 4]

    if n > 3:
        pts.append(rd(n + 1, [(u2, n - 1), (v2, 2)]))
        k_range.append(5)

    b0 = ndim.nsimplex.integrate_monomial(n * [0], symbolic=True)
    b = [
        ndim.nsimplex.integrate_monomial(np.array([k] + (n - 1) * [0]), symbolic=True)
        / b0
        for k in k_range
    ]

    A = [[sum(p[:, 0] ** k) for p in pts] for k in k_range]

    flt = np.vectorize(float)
    w = np.linalg.solve(flt(A), flt(b))

    data = [(w[i], pts[i]) for i in range(len(w))]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return TnScheme("Stroud 1969", n, weights, points, degree, source)



import warnings

from sympy import Rational as frac
from sympy import sqrt

from ..helpers import book
from ._helpers import E2r2Scheme, register
from ._rabinowitz_richter import rabinowitz_richter_1 as stroud_9_1
from ._rabinowitz_richter import rabinowitz_richter_2 as stroud_11_1
from ._rabinowitz_richter import rabinowitz_richter_3 as stroud_11_2
from ._rabinowitz_richter import rabinowitz_richter_4 as stroud_13_1
from ._rabinowitz_richter import rabinowitz_richter_5 as stroud_15_1
from ._stroud_secrest import stroud_secrest_5 as stroud_5_1
from ._stroud_secrest import stroud_secrest_6 as stroud_7_1

_source = book(
    authors=["Arthur Stroud"],
    title="Approximate Calculation of Multiple Integrals",
    publisher="Prentice Hall",
    year="1971",
)


def stroud_4_1():
    d = {
        "zero2": [[frac(1, 2)]],
        "d5.0": [[frac(1, 10)], [[sqrt(2)]]],
    }
    return E2r2Scheme("Stroud 4-1", d, 4, _source, 2.220e-16)


def stroud_5_2():
    # Cartesian product Gauss formula
    r = sqrt(frac(3, 2))
    d = {
        "zero2": [[frac(4, 9)]],
        "d4_a0": [[frac(1, 9)], [r]],
        "d4_aa": [[frac(1, 36)], [r]],
    }
    return E2r2Scheme("Stroud 5-2", d, 5, _source, 4.441e-16)


def stroud_7_2():
    # Cartesian product Gauss formula
    sqrt6 = sqrt(6)
    r, s = (sqrt((3 + p_m * sqrt6) / 2) for p_m in [+1, -1])
    A, B = ((5 - p_m * 2 * sqrt6) / 48 for p_m in [+1, -1])
    C = frac(1, 48)

    d = {"d4_a0": [[A, B], [r, s]], "d4_ab": [[C], [r], [s]]}

    # TODO find what's wrong
    warnings.warn("Stroud's Gauss product formula has degree 1, not 7.")
    return E2r2Scheme("Stroud 7-2", d, 1, _source, 2.220e-16)


register(
    [
        stroud_4_1,
        stroud_5_1,
        stroud_5_2,
        stroud_7_1,
        stroud_7_2,
        stroud_9_1,
        stroud_11_1,
        stroud_11_2,
        stroud_13_1,
        stroud_15_1,
    ]
)



from ._class import QuadratureScheme
from .combinatorics import combine, fs_array, fsd, get_all_exponents, pm, pm_roll, rd, z
from .misc import (
    article,
    book,
    comb,
    compute_dobrodeev,
    gamma_n_2,
    get_nsimplex_points,
    n_outer,
    online,
    phdthesis,
    prod,
    techreport,
    untangle,
)
from .plot import backend_to_function, plot_disks, plot_disks_1d, show_mpl, show_vtk
from .symmetries import expand_symmetries, expand_symmetries_points_only

__all__ = [
    "QuadratureScheme",
    "z",
    "rd",
    "fsd",
    "fs_array",
    "combine",
    "pm",
    "pm_roll",
    "gamma_n_2",
    "get_all_exponents",
    "untangle",
    "n_outer",
    "comb",
    "compute_dobrodeev",
    "get_nsimplex_points",
    "article",
    "book",
    "techreport",
    "phdthesis",
    "prod",
    "online",
    "plot_disks_1d",
    "plot_disks",
    "show_mpl",
    "show_vtk",
    "backend_to_function",
    "expand_symmetries",
    "expand_symmetries_points_only",
]



import numpy as np
import orthopy

from ..tools import scheme_from_rc
from ._helpers import C1Scheme


def gauss_jacobi(n, alpha, beta, mode="numpy"):
    degree = 2 * n - 1

    rc = orthopy.c1.jacobi.RecurrenceCoefficients("monic", alpha, beta, symbolic=True)
    _, a, b = np.array([rc[k] for k in range(n)]).T
    points, weights = scheme_from_rc(a, b, rc.int_1, mode=mode)
    return C1Scheme("Gauss-Jacobi", degree, weights, points)



import numpy as np

from ._helpers import C2Scheme


def product(scheme1d):
    schemes = scheme1d if isinstance(scheme1d, list) else 2 * [scheme1d]

    weights = np.outer(schemes[0].weights, schemes[1].weights).flatten()
    assert len(weights) > 0
    points = np.array(np.meshgrid(schemes[0].points, schemes[1].points)).reshape(2, -1)
    weights /= 4
    degree = min(s.degree for s in schemes)
    return C2Scheme(
        f"Product scheme ({scheme1d.name})",
        {"plain": np.vstack([weights, points])},
        degree,
    )



import json
import math

import numpy as np

from ..helpers import QuadratureScheme, expand_symmetries, plot_disks

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class E2r2Scheme(QuadratureScheme):
    def __init__(
        self, name, symmetry_data, degree, source, tol=1.0e-14, weight_factor=None
    ):
        self.domain = "E2r2"
        self.symmetry_data = symmetry_data
        points, weights = expand_symmetries(symmetry_data, dim=2)
        assert points.shape[0] == 2
        if weight_factor is not None:
            weights *= weight_factor
        super().__init__(name, weights, points, degree, source, tol)

    def plot(self, show_axes=False):
        from matplotlib import pyplot as plt

        ax = plt.gca()
        plt.axis("equal")

        if not show_axes:
            ax.set_axis_off()

        I0 = 2 * math.pi
        plot_disks(plt, self.points.T, self.weights, I0)

    def integrate(self, f, dot=np.dot):
        flt = np.vectorize(float)
        ref_vol = math.pi
        return ref_vol * dot(f(flt(self.points)), flt(self.weights))


def _read(filepath, source):
    with open(filepath) as f:
        content = json.load(f)

    degree = content["degree"]
    name = content["name"]
    tol = content["test_tolerance"]
    data = content["data"]

    weight_factor = content["weight factor"] if "weight factor" in content else None

    return E2r2Scheme(name, data, degree, source, tol, weight_factor=weight_factor)


def get_good_scheme(degree):
    if degree <= 15:
        return {
            0: schemes["stroud_4_1"],
            1: schemes["stroud_4_1"],
            2: schemes["stroud_4_1"],
            3: schemes["stroud_4_1"],
            4: schemes["stroud_4_1"],
            5: schemes["stroud_secrest_5"],
            6: schemes["stroud_secrest_6"],
            7: schemes["stroud_secrest_6"],
            8: schemes["haegemans_piessens_a"],
            9: schemes["haegemans_piessens_a"],
            10: schemes["rabinowitz_richter_2"],
            11: schemes["rabinowitz_richter_2"],
            12: schemes["rabinowitz_richter_5"],
            13: schemes["rabinowitz_richter_5"],
            14: schemes["rabinowitz_richter_5"],
            15: schemes["rabinowitz_richter_5"],
        }[degree]()

    return None



from ._dobrodeev_1978 import dobrodeev_1978
from ._mysovskikh import mysovskikh_1, mysovskikh_2
from ._stroud import (
    stroud_un_3_1,
    stroud_un_3_2,
    stroud_un_5_1,
    stroud_un_5_2,
    stroud_un_5_3,
    stroud_un_5_4,
    stroud_un_7_1,
    stroud_un_7_2,
    stroud_un_11_1,
)
from ._stroud_1967 import stroud_1967
from ._stroud_1969 import stroud_1969

__all__ = [
    "dobrodeev_1978",
    "mysovskikh_1",
    "mysovskikh_2",
    "stroud_un_3_1",
    "stroud_un_3_2",
    "stroud_un_5_1",
    "stroud_un_5_2",
    "stroud_un_5_3",
    "stroud_un_5_4",
    "stroud_un_7_2",
    "stroud_un_7_1",
    "stroud_un_11_1",
    "stroud_1967",
    "stroud_1969",
]



import sys


def optimize(content):
    if "domain" not in content:
        print('Missing key "domain".', file=sys.stderr)
        exit(1)

    domain = content["domain"]
    if domain.lower() == "t2":
        return _optimize_t2(content)
    elif domain.lower() == "s2":
        return _optimize_s2(content)
    elif domain.lower() == "c2":
        return _optimize_c2(content)
    elif domain.lower() == "u3":
        return _optimize_u3(content)

    print(f'Don\'t know how to optimize domain "{domain}".', file=sys.stderr)
    exit(1)


def _optimize_u3(content):
    import orthopy

    from .u3._helpers import (
        _scheme_from_dict,
        expand_symmetries,
        expand_symmetries_points_only,
    )

    return _optimize(
        content,
        expand_symmetries,
        expand_symmetries_points_only,
        _scheme_from_dict,
        get_evaluator=lambda points: orthopy.u3.EvalCartesian(
            points, scaling="quantum mechanic"
        ),
    )


def _optimize_c2(content):
    import orthopy

    from .c2._helpers import (
        _scheme_from_dict,
        expand_symmetries,
        expand_symmetries_points_only,
    )

    return _optimize(
        content,
        expand_symmetries,
        expand_symmetries_points_only,
        _scheme_from_dict,
        get_evaluator=lambda points: orthopy.cn.Eval(points),
    )


def _optimize_s2(content):
    import orthopy

    from .s2._helpers import (
        _scheme_from_dict,
        expand_symmetries,
        expand_symmetries_points_only,
    )

    return _optimize(
        content,
        expand_symmetries,
        expand_symmetries_points_only,
        _scheme_from_dict,
        get_evaluator=lambda points: orthopy.s2.zernike.Eval(points, scaling="normal"),
    )


def _optimize_t2(content):
    import orthopy

    from .helpers import expand_symmetries, expand_symmetries_points_only
    from .t2._helpers import _scheme_from_dict

    # return _optimize_weights_as_variables(
    return _optimize(
        content,
        expand_symmetries,
        expand_symmetries_points_only,
        _scheme_from_dict,
        get_evaluator=lambda points: orthopy.t2.Eval(points, scaling="normal"),
    )


def _optimize(
    content,
    expand_symmetries,
    expand_symmetries_points_only,
    scheme_from_dict,
    get_evaluator,
):
    """Compute the weights from the points via a least-squares problem. Only the point
    coordinates are variables.
    """
    import numpy as np
    from scipy.optimize import minimize

    degree = content["degree"]
    keys = list(content["data"].keys())
    num_symm = [len(item[0]) for item in content["data"].values()]

    def x_to_dict(x):
        # convert x to dictionary (without weights)
        x_split = np.split(x, splits)
        vals = [item.reshape(shape) for item, shape in zip(x_split, shapes)]
        d = dict(zip(keys, vals))
        return d

    def get_w_from_x(x):
        d = x_to_dict(x)
        points, len_symm = expand_symmetries_points_only(d, dim=2)

        if np.any(np.isnan(points)):
            # return some "large" residual value
            return None, None, None, 1.0

        # evaluate all orthogonal polynomials up to `degree` at all points
        evaluator = get_evaluator(points)
        A2 = np.concatenate([next(evaluator) for _ in range(degree + 1)])

        assert sum(a * b for a, b in zip(len_symm, num_symm)) == A2.shape[1]

        # sum up all columns which belong to the same weight
        k = 0
        sums = []
        for lsym, nsym in zip(len_symm, num_symm):
            for i in range(nsym):
                sums.append(np.sum(A2[:, k + i : k + lsym * nsym : nsym], axis=1))
            k += lsym * nsym

        A = np.column_stack(sums)

        # The exact values are 0 except for the first entry
        b = np.zeros(A.shape[0])
        b[0] = evaluator.int_p0
        # b[0] /= np.pi  # necessary for S2

        w, res, rank, s = np.linalg.lstsq(A, b, rcond=None)

        # spherical harmonics (for u3) are complex-valued
        assert np.all(np.abs(w.imag) < 1.0e-15)
        w = w.real

        if rank < min(A.shape):
            # return some "large" residual value
            return None, None, None, 1.0

        return A, b, w, np.sqrt(res[0])

    def f(x):
        _, _, _, res = get_w_from_x(x)
        # print(f"f(x) = {res:.6e}")
        return res

    keys = list(content["data"].keys())
    values = list(content["data"].values())

    values_without_weights = [val[1:] for val in values]
    shapes = [np.array(val).shape for val in values_without_weights]
    sizes = [np.array(val).size for val in values_without_weights]
    splits = np.cumsum(sizes)[:-1]

    # collect and concatenate all point coords
    x0 = np.concatenate([np.array(val).flat for val in values_without_weights])

    # TODO check initial residual with original weights
    r0 = f(x0)

    out = minimize(f, x0, method="Nelder-Mead", tol=1.0e-17)

    # Don't fail on `not out.success`. It could be because of
    # ```
    # Maximum number of function evaluations has been exceeded
    # ```
    # but the scheme could still have improved.

    # out.fun == f(out.x) exactly
    if r0 <= out.fun:
        raise RuntimeError("Optimization couldn't improve scheme.")

    # compute max(err)
    A, b, w, _ = get_w_from_x(out.x)
    # assert A is not None
    # max_res = np.max(np.abs(A @ w - b))

    # Compute max_res exactly like in the tests
    d = x_to_dict(out.x)
    # prepend weights
    k = 0
    for key, value in d.items():
        if len(value.shape) == 1:
            n = 1
            d[key] = [[w[k]]]
        else:
            n = value.shape[1]
            d[key] = np.column_stack([w[k : k + n], value.T]).T
        k += n
    # content["data"] = d
    # scheme = scheme_from_dict(content)
    # max_res = max(scheme.compute_residuals(degree))

    # print(max_res)

    return d, out.fun, np.linalg.cond(A)


def _optimize_weights_as_variables(
    content,
    expand_symmetries,
    expand_symmetries_points_only,
    scheme_from_dict,
    get_evaluator,
):
    """Treat the weights as variables.

    It turns out from numerical experiments that this is inferior to the above approach.
    TODO Information about the derivative + using Newton could improve it a lot.
    """
    import numpy as np
    from scipy.optimize import minimize

    degree = content["degree"]
    keys = list(content["data"].keys())
    A = None

    def x_to_dict(x):
        # convert x to dictionary (without weights)
        x_split = np.split(x, splits)
        vals = [item.reshape(shape) for item, shape in zip(x_split, shapes)]
        d = dict(zip(keys, vals))
        return d

    def get_system(x):
        d = x_to_dict(x)
        points, weights = expand_symmetries(d)

        if np.any(np.isnan(points)):
            # return some "large" residual value
            return 1.0

        # evaluate all orthogonal polynomials up to `degree` at all points
        evaluator = get_evaluator(points.T)
        A = np.concatenate([next(evaluator) for _ in range(degree + 1)])

        # The exact values are 0 except for the first entry
        b = np.zeros(A.shape[0])
        b[0] = evaluator.int_p0
        return A, weights, b

    def f(x):
        A, weights, b = get_system(x)
        res = A @ weights - b
        res_norm = np.sqrt(np.dot(res, res))
        return res_norm

    keys = list(content["data"].keys())
    values = list(content["data"].values())

    shapes = [np.array(val).shape for val in values]
    sizes = [np.array(val).size for val in values]
    splits = np.cumsum(sizes)[:-1]

    # collect and concatenate all point coords
    x0 = np.concatenate([np.array(val).flat for val in values])

    # TODO check initial residual with original weights
    r0 = f(x0)

    out = minimize(f, x0, method="Nelder-Mead", tol=1.0e-17)

    # Don't fail on `not out.success`. It could be because of
    # ```
    # Maximum number of function evaluations has been exceeded
    # ```
    # but the scheme could still have improved.

    # out.fun == f(out.x) exactly
    if r0 <= out.fun:
        raise RuntimeError("Optimization couldn't improve scheme.")

    d = x_to_dict(out.x)

    A, _, _ = get_system(out.x)

    return d, out.fun, np.linalg.cond(A)



import math

import numpy as np
import sympy

from ..helpers import article, untangle
from ._helpers import SnScheme

source = article(
    authors=["A.H. Stroud"],
    title="Remarks on the Disposition of Points in Numerical Integration Formulas",
    journal="Mathematical Tables and Other Aids to Computation",
    volume="11",
    number="60",
    month="oct",
    year="1957",
    pages="257-261",
    url="https://doi.org/10.2307/2001945",
)


def stroud_1957(n, symbolic=False):
    frac = sympy.Rational if symbolic else lambda a, b: a / b
    sqrt = sympy.sqrt if symbolic else math.sqrt
    sin = sympy.sin if symbolic else math.sin
    cos = sympy.cos if symbolic else math.cos
    pi = sympy.pi if symbolic else math.pi

    pts = [
        [
            [
                sqrt(frac(2, n + 2)) * cos(2 * k * i * pi / (n + 1))
                for i in range(n + 1)
            ],
            [
                sqrt(frac(2, n + 2)) * sin(2 * k * i * pi / (n + 1))
                for i in range(n + 1)
            ],
        ]
        for k in range(1, n // 2 + 1)
    ]
    if n % 2 == 1:
        sqrt3pm = np.full(n + 1, 1 / sqrt(n + 2))
        sqrt3pm[1::2] *= -1
        pts.append(sqrt3pm)
    pts = np.vstack(pts).T

    data = [(frac(1, n + 1), pts)]

    points, weights = untangle(data)
    points = np.ascontiguousarray(points.T)
    return SnScheme("Stroud 1957", n, weights, points, 2, source)



import numpy as np
import sympy

from .. import helpers
from ..cn import CnScheme
from ..cn import ncube_points as cube_points
from ..cn import transform

schemes = {}


def register(in_schemes):
    for scheme in in_schemes:
        schemes[scheme.__name__] = scheme


class C3Scheme(CnScheme):
    def __init__(self, name, symmetry_data, degree, source=None, tol=1.0e-14):
        self.symmetry_data = symmetry_data
        points, weights = helpers.expand_symmetries(symmetry_data, dim=3)
        super().__init__(name, 3, weights, points, degree, source, tol)
        self.domain = "C3"

    def show(self, hexa=cube_points([0.0, 1.0], [0.0, 1.0], [0.0, 1.0]), backend="vtk"):
        """Shows the quadrature points on a given hexahedron. The size of the balls
        around the points coincides with their weights."""
        edges = np.array(
            [
                [hexa[0, 0, 0], hexa[1, 0, 0]],
                [hexa[1, 0, 0], hexa[1, 1, 0]],
                [hexa[1, 1, 0], hexa[0, 1, 0]],
                [hexa[0, 1, 0], hexa[0, 0, 0]],
                #
                [hexa[0, 0, 1], hexa[1, 0, 1]],
                [hexa[1, 0, 1], hexa[1, 1, 1]],
                [hexa[1, 1, 1], hexa[0, 1, 1]],
                [hexa[0, 1, 1], hexa[0, 0, 1]],
                #
                [hexa[0, 0, 0], hexa[0, 0, 1]],
                [hexa[1, 0, 0], hexa[1, 0, 1]],
                [hexa[1, 1, 0], hexa[1, 1, 1]],
                [hexa[0, 1, 0], hexa[0, 1, 1]],
            ]
        )
        edges = np.moveaxis(edges, 1, 2)

        helpers.backend_to_function[backend](
            transform(self.points, hexa),
            self.weights,
            self.integrate(lambda x: np.ones(x.shape[1:]), hexa),
            edges,
        )


def get_good_scheme(degree):
    if degree <= 11:
        return {
            0: schemes["midpoint"](),
            1: schemes["midpoint"](),
            2: schemes["face_midpoint"](),
            3: schemes["face_midpoint"](),
            4: schemes["hammer_stroud_4_3"](),
            5: schemes["hammer_stroud_4_3"](),
            6: schemes["hammer_wymore"](sympy.Rational(27, 20)),
            7: schemes["hammer_wymore"](sympy.Rational(27, 20)),
            8: schemes["witherden_vincent_09"](),
            9: schemes["witherden_vincent_09"](),
            10: schemes["witherden_vincent_11"](),
            11: schemes["witherden_vincent_11"](),
        }[degree]

    return None



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import C3Scheme, register

source = article(
    authors=["A.H. Stroud"],
    title="Some fifth degree integration formulas for symmetric regions II",
    journal="Numerische Mathematik",
    volume="9",
    number="5",
    month="apr",
    year="1967",
    pages="460-468",
    url="https://doi.org/10.1007/BF02162160",
)


def stroud_1967():
    # ENH Analytic expression for all quantities are given in
    #
    # J.W. Peterson,
    # Analytical Formulae for Two of A. H. Stroud's Quadrature Rules,
    # Sep. 2009,
    # <https://arxiv.org/pdf/0909.5106.pdf>.
    sqrt19 = sqrt(19)
    t = sqrt(71440 + 6802 * sqrt19)

    lmbd, gmma = (sqrt((1919 - 148 * sqrt19 + i * 4 * t) / 3285) for i in [+1, -1])
    xi, mu = (-sqrt((1121 + 74 * sqrt19 - i * 2 * t) / 3285) for i in [+1, -1])
    mu *= -1

    B, C = (
        133225 / (260072 - 1520 * sqrt19 + i * (133 - 37 * sqrt19) * t)
        for i in [+1, -1]
    )

    d = {
        "zero3": [[frac(32, 19) / 8]],
        "symm_rss_pm": [[B / 8, C / 8], [lmbd, gmma], [xi, mu]],
    }
    return C3Scheme("Stroud 1967", d, 5, source)


register([stroud_1967])



from sympy import Rational as frac
from sympy import sqrt

from ..helpers import article
from ._helpers import E2rScheme, register

_source = article(
    authors=["A.H. Stroud", "D. Secrest"],
    title="Approximate integration formulas for certain spherically symmetric regions",
    journal="Math. Comp.",
    volume="17",
    year="1963",
    pages="105-135",
    url="https://doi.org/10.1090/S0025-5718-1963-0161473-0",
)


def stroud_secrest_5():
    nu = 2 * sqrt(5)
    xi = sqrt(5)
    eta = sqrt(15)

    d = {
        "zero2": [[frac(7, 10)]],
        "c2_a0": [[frac(1, 20)], [nu]],
        "sxy": [[frac(1, 20)], [xi], [eta]],
    }
    return E2rScheme("Stroud-Secrest V", d, 5, _source)


def stroud_secrest_6():
    sqrt74255 = sqrt(74255)

    nu = sqrt(42)
    xi, eta = (sqrt((6615 - p_m * 21 * sqrt74255) / 454) for p_m in [+1, -1])
    A = frac(5, 588)
    B, C = ((5272105 + p_m * 18733 * sqrt74255) / 43661940 for p_m in [+1, -1])

    d = {
        "d4_a0": [[A], [nu]],
        "d4_aa": [[B, C], [xi, eta]],
    }
    return E2rScheme("Stroud-Secrest VI", d, 7, _source)


register([stroud_secrest_5, stroud_secrest_6])



import numpy as np
import sympy
from mpmath import mp

from ._helpers import C1Scheme


def chebyshev_gauss_1(n, mode="numpy"):
    """Chebyshev-Gauss quadrature for \\int_{-1}^1 f(x) / sqrt(1+x^2) dx."""
    degree = n if n % 2 == 1 else n + 1

    if mode == "numpy":
        points = np.cos((2 * np.arange(1, n + 1) - 1) / (2 * n) * np.pi)
        weights = np.full(n, np.pi / n)
    elif mode == "sympy":
        points = np.array(
            [
                sympy.cos(sympy.Rational(2 * k - 1, 2 * n) * sympy.pi)
                for k in range(1, n + 1)
            ]
        )
        weights = np.full(n, sympy.pi / n)
    else:
        assert mode == "mpmath"
        points = np.array(
            [mp.cos(mp.mpf(2 * k - 1) / (2 * n) * mp.pi) for k in range(1, n + 1)]
        )
        weights = np.full(n, mp.pi / n)
    return C1Scheme("Chebyshev-Gauss 1", degree, weights, points)


def chebyshev_gauss_2(n, mode="numpy", decimal_places=None):
    """Chebyshev-Gauss quadrature for \\int_{-1}^1 f(x) * sqrt(1+x^2) dx."""
    degree = n if n % 2 == 1 else n + 1

    # TODO make explicit for all modes
    if mode == "numpy":
        points = np.cos(np.pi * np.arange(1, n + 1) / (n + 1))
        weights = np.pi / (n + 1) * (np.sin(np.pi * np.arange(1, n + 1) / (n + 1))) ** 2
    elif mode == "sympy":
        points = np.array(
            [sympy.cos(sympy.Rational(k, n + 1) * sympy.pi) for k in range(1, n + 1)]
        )
        weights = np.array(
            [
                sympy.pi / (n + 1) * sympy.sin(sympy.pi * sympy.Rational(k, n + 1)) ** 2
                for k in range(1, n + 1)
            ]
        )
    else:
        assert mode == "mpmath"
        points = np.array(
            [mp.cos(mp.mpf(k) / (n + 1) * mp.pi) for k in range(1, n + 1)]
        )
        weights = np.array(
            [
                mp.pi / (n + 1) * mp.sin(mp.pi * mp.mpf(k) / (n + 1)) ** 2
                for k in range(1, n + 1)
            ]
        )
    return C1Scheme("Chebyshev-Gauss 2", degree, weights, points)



import pathlib

from ...helpers import article
from .._helpers import _read, register

_source = article(
    authors=["Robert Piessens", "Ann Haegemans"],
    title="Cubature Formulas of Degree Nine for Symmetric Planar Regions",
    journal="Mathematics of Computation",
    volume="29",
    number="11",
    month="jul",
    year="1975",
    pages="810-815",
    url="https://doi.org/10.1090/S0025-5718-1975-0368393-5",
)

this_dir = pathlib.Path(__file__).resolve().parent


def piessens_haegemans():
    return _read(this_dir / "piessens_haegemans.json", _source)


register([piessens_haegemans])
